# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module random_variable.

'''

from copy import deepcopy
import numpy as np
import tensap


class RandomVariable(tensap.ProbabilityMeasure):
    '''
    Class RandomVariable.

    Attributes
    ----------
    moments : numpy.array
        The moments of the normal random variable (if computed).

    '''

    def __init__(self):
        '''
        Constructor of the class RandomVariable.

        The moments attribute remains empty as long as the moments have not
        been computed using the method moment.

        Returns
        -------
        None.

        '''
        self.moments = np.array([])

    @staticmethod
    def ndim():
        '''
        Return the dimension of the random variable, equal to 1.

        Returns
        -------
        int
            The dimension of the random variable.

        '''
        return 1

    def __eq__(self, rv_2):
        if not (isinstance(self, tensap.RandomVariable) and
                isinstance(rv_2, tensap.RandomVariable)):
            is_equal = False
        elif not isinstance(self, type(rv_2)):
            is_equal = False
        else:
            is_equal = True
            param_1 = self.get_parameters()
            param_2 = rv_2.get_parameters()
            for ind in zip(param_1, param_2):
                is_equal = is_equal and (ind[0] == ind[1])
        return is_equal

    def __neq__(self, rv_2):
        return not (self == rv_2)

    def number_of_parameters(self):
        '''
        Compute the number of parameters that admits the random variable.

        Returns
        -------
        int
            The number of parameters that admits the random variable.

        '''
        return np.size(self.get_parameters())

    def pdf(self, x):
        '''
        Compute the probability density function (pdf) of the RandomVariable
        at points x.

        Parameters
        ----------
        x : float or list or numpy.ndarray
            The points at which the pdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the pdf at points x.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def cdf(self, x):
        '''
        Compute the cumulative distribution function (cdf) of the
        RandomVariable at points x.

        Parameters
        ----------
        x : float or list or numpy.ndarray
            The points at which the cdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the cdf at points x.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def icdf(self, x):
        '''
        Compute the inverse cumulative distribution function (icdf) of the
        RandomVariable at points x.

        Parameters
        ----------
        x : float or list or numpy.ndarray
            The points at which the icdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the icdf at points x.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def iso_probabilistic_grid(self, n):
        '''
        Return a set of n+1 points (x_0, ..., x_{n}) such that the n sets
        (x0, x_1), [x_1, x_2) ... [x_{n-1}, x_{n})  have all the same
        probability p = 1/n (with x0 = self.min() and x_{n+1}=self.max()).

        Parameters
        ----------
        n : int
            The number of points of the grid plus one.

        Returns
        -------
        numpy.ndarray
            The iso-probabilistic grid.

        '''
        if n < 1:
            n = int(np.ceil(1/n))

        if n == 1:
            g = []
        elif n == 2:
            g = [self.icdf(0.5)]
        else:
            g = self.icdf(np.linspace(1/n, 1-1/n, n-1))
        return np.concatenate(([self.min()], g, [self.max()]))

    def discretize(self, n):
        '''
        Return a discrete random variable taking n possible values x1, ..., xn,
        these values being the quantiles of self of probability 1/(2n) + i/n,
        i=0n ..., n-1 and such that P(Xn >= xn) = 1/n.

        Parameters
        ----------
        n : int
            The number of possible values the discrete random variable can
            take.

        Returns
        -------
        tensap.DiscreteRandomVariable
            The obtained discrete random variable.

        '''
        u = np.linspace(1/(2*n), 1-1/(2*n), n)
        x = self.icdf(u)
        return tensap.DiscreteRandomVariable(x)

    def gauss_integration_rule(self, nb_pts):
        '''
        Return the nb_pts-points gauss integration rule associated with the
        measure of self, using Golub-Welsch algorithm.

        Parameters
        ----------
        nb_pts : int
            The number of integration points.

        Returns
        -------
        tensap.IntegrationRule
            The integration rule associated with the measure of self.

        '''
        poly = self.orthonormal_polynomials(nb_pts+1)
        if isinstance(poly, tensap.ShiftedOrthonormalPolynomials):
            shift = poly.shift
            scaling = poly.scaling
            poly = poly.polynomials
            flag = True
        else:
            flag = False

        coef = poly._recurrence_coefficients
        if coef.shape[1] < nb_pts:
            coef = poly.recurrence(poly.measure, nb_pts-1)
        else:
            coef = coef[:, :nb_pts]

        # Jacobi matrix
        if nb_pts == 1:
            jacobi_matrix = np.diag(coef[0, :])
        else:
            jacobi_matrix = np.diag(coef[0, :]) + \
                np.diag(np.sqrt(coef[1, 1:]), -1) + \
                np.diag(np.sqrt(coef[1, 1:]), 1)

        # Quadrature points are the eigenvalues of the Jacobi matrix, weights
        # are deduced from the eigenvectors
        eig_values, eig_vectors = np.linalg.eig(jacobi_matrix)
        points = np.sort(eig_values)
        ind = np.argsort(eig_values)
        eig_vectors = eig_vectors[:, ind]

        weights = eig_vectors[0, :]**2 / np.sqrt(np.sum(eig_vectors**2, 0))

        if flag:
            points = shift + scaling * points
        return tensap.IntegrationRule(points, weights)

    def lhs_random(self, n, p=1):
        '''
        Latin Hypercube Sampling of the random variable self of n points in
        dimension p.

        Requires the package pyDOE.

        Parameters
        ----------
        n : int
            Number of points.
        p : int, optional
            The dimension. The default is 1.

        Returns
        -------
        numpy.ndarray
            The coordinates of the Latin Hypercube Sampling in each dimension.

        '''
        from pyDOE import lhs
        A = lhs(p, samples=n)
        U = tensap.UniformRandomVariable(0, 1)
        A = [U.transfer(self, A[:, i]) for i in range(A.shape[1])]
        return np.transpose(np.array(A))

    def likelihood(self, x):
        '''
         Compute the log-likelihood of the random variable on sample x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The sample used to compute the log-likelihood.

        Returns
        -------
        float
            The log-likelihood of the random variable on sample x.

        '''
        P = self.pdf(x)
        return np.sum(np.log(P + np.finfo(float).eps))

    def max(self):
        '''
        Compute the maximum value that can take the inverse cumulative
        distribution function of the random variable.

        Returns
        -------
        float
            The maximum value that can take the inverse cumulative distribution
            function of the random variable.

        '''
        return np.max(self.support())

    def min(self):
        '''
        Compute the minimum value that can take the inverse cumulative
        distribution function of the random variable.

        Returns
        -------
        float
            The minimum value that can take the inverse cumulative distribution
            function of the random variable.

        '''
        return np.min(self.support())

    def moment(self, ind, nargout=1):
        '''
        Compute the moments of self of orders contained in ind, defined as
        E(X^ind[î]).
        If a second output argument is asked, the computed moments are stored
        in the random variable X.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The orders of the moments.
        nargout : int, optional
            Indicates the number of expected outputs. The default is 1,
            indicating to return only the moments.

        Returns
        -------
        numpy.ndarray
            The computed moments.
        tensap.RandomVariable
            The RandomVariable object with the computed moments stored in the
            attribute moments.

        '''
        ind = np.atleast_1d(ind)
        if np.size(self.moments)-1 >= np.max(ind):
            out = self.moments[ind]
        else:
            out = np.zeros(np.size(ind))
            nb_pts = int(np.ceil((np.max(ind)+1)/2))
            G = self.gauss_integration_rule(nb_pts)
            for nb, ind_loc in enumerate(ind):
                out[nb] = G.integrate(lambda x: x ** ind_loc)

        if nargout == 1:
            return out
        X = deepcopy(self)
        X.moments = out
        return out, X

    def mean(self):
        '''
        Return the mean of the random variable.

        Returns
        -------
        float
            The mean of the random variable.

        '''
        return self.random_variable_statistics()[0]

    def std(self):
        '''
        Return the standard deviation of the random variable.

        Returns
        -------
        float
            The standard deviation of the random variable.

        '''
        return np.sqrt(self.random_variable_statistics()[1])

    def variance(self):
        '''
        Return the variance of the random variable.

        Returns
        -------
        float
            The variance of the random variable.

        '''
        return self.random_variable_statistics()[1]

    def transfer(self, Y, x):
        '''
        Transfer from the random variable self to the random variable Y at
        points x.

        Parameters
        ----------
        Y : tensap.RandomVariable
            The target RandomVariable of the transfer.
        x : list or numpy.ndarray
            The input points.

        Returns
        -------
        y : numpy.ndarray
            The transfered points.

        '''
        assert isinstance(self, tensap.RandomVariable) and \
            isinstance(Y, tensap.RandomVariable), \
            'The first two arguments must be RandomVariable.'
        return Y.icdf(self.cdf(x))

    def truncated_support(self):
        '''
        Return the truncated support of the random variable.

        Returns
        -------
        sup : numpy.ndarray
            The truncated support of the random variable.

        '''

        sup = self.support()
        if sup[0] == -np.inf:
            sup[0] = self.mean() - 10*self.std()
        if sup[1] == np.inf:
            sup[1] = self.mean() + 10*self.std()
        return sup

    def pdf_plot(self, *args):
        '''
        Plot the probability density function (pdf) of the random variable.

        See also plot.

        Parameters
        ----------
        *args : tuple
            Additional parameters of the method plot.

        Returns
        -------
        None.

        '''
        self.plot('pdf', *args)

    def cdf_plot(self, *args):
        '''
        Plot the cumulative distribution function (cdf) of the random variable.

        See also plot.

        Parameters
        ----------
        *args : tuple
            Additional parameters of the method plot.

        Returns
        -------
        None.

        '''
        self.plot('cdf', *args)

    def icdf_plot(self, *args):
        '''
        Plot the inverse cumulative distribution function (icdf) of the random
        variable.

        See also plot.

        Parameters
        ----------
        *args : tuple
            Additional parameters of the method plot.

        Returns
        -------
        None.

        '''
        self.plot('icdf', *args)

    def plot(self, quantity, n_pts=100, bar=False, *args):
        '''
        Plot the desired quantity, chosen between 'pdf', 'cdf' or 'icdf'.

        Parameters
        ----------
        quantity : str
            The desired quantity, chosen between 'pdf', 'cdf' or 'icdf'.
        n_pts : int, optional
            The number of points used for the plot. The default is 100.
        bar : boolean, optional
            Determines if the method uses matplotlib.pyplot's function bar
            or plot. The default is False.
        *args : tuple
            Additional parameters for matplotlib.pyplot's function plot or bar.

        Raises
        ------
        ValueError
            If the provided argument quantity is wrong.

        Returns
        -------
        None.

        '''
        import matplotlib.pyplot as plt

        sup = self.truncated_support()
        if quantity == 'cdf':
            x = np.linspace(sup[0], sup[1], n_pts)
            P = self.cdf(x)
        elif quantity == 'icdf':
            x = np.linspace(0, 1, n_pts)
            P = self.icdf(x)
        elif quantity == 'pdf':
            x = np.linspace(sup[0], sup[1], n_pts)
            P = self.pdf(x)
        else:
            raise ValueError('Wrong argument value')

        if bar:
            plt.bar(x, P, *args)
        else:
            plt.plot(x, P, *args)
        plt.show()

    def get_parameters(self):
        '''
        Return the parameters of the random variable.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def random_variable_statistics(self):
        '''
        Return the mean and the variance of the random variable.

        Returns
        -------
        float
            The mean of the random variable.
        float
            The variance of the random variable.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def random(self, n):
        '''
        Generate n random numbers according to the distribution of the
        RandomVariable.

        Parameters
        ----------
        n : int
            The number of random numbers generated.

        Returns
        -------
        numpy.ndarray
            The generated numbers.

        '''
        raise NotImplementedError('No generic implementation of the method.')

    def orthonormal_polynomials(self, *max_degree):
        '''
        Return the max_degree-1 first orthonormal polynomials associated with
        the RandomVariable.

        Parameters
        ----------
        max_degree : int, optional
            The maximum degree of the returned polynomials. The default is
            None, choosing the default maximum degree associated with the
            constructor of the polynomials.

        Returns
        -------
        poly : tensap.OrthonormalPolynomials
            The generated orthonormal polynomials.

        '''
        raise NotImplementedError('No generic implementation of the method.')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module grids.

'''

from abc import ABC, abstractmethod
import numpy as np
import tensap


class TensorGrid(ABC):
    '''
    Class TensorGrid: grids in product sets.

    Attributes
    ----------
    dim : int
        The dimension.
    grids: list
        List containing the grids.
    shape: numpy.ndarray
        The shapes of the grids.

    '''

    def __init__(self):
        '''
        Constructor for the class TensorGrid.

        Returns
        -------
        None.

        '''
        self.dim = None
        self.grids = None
        self.shape = None

    def ndim(self):
        '''
        Return the dimension of the underlying space.

        Returns
        -------
        int
            The dimension of the underlying space.

        '''
        return self.dim

    def shape(self, d=None):
        '''
        Return the shape of the grid, along the dimension d if provided.

        Equivalent to self.shape[d] if d is provided, self.shape otherwise.

        Parameters
        ----------
        d : int, optional
            The dimension for which the shape is asked. The default is None.

        Returns
        -------
        numpy.ndarray
            The shape of the grid.

        '''
        if d is None:
            return self.shape
        return self.shape[d]

    def plot_grid(self, *args, **kwargs):
        '''
        Plot the grid.

        Parameters
        ----------
        *args : tuple
            Parameters of the plot, see matplotlib.pyplot.plot for more
            information.

        Raises
        ------
        ValueError
            If the dimension is greater than 3.

        Returns
        -------
        None.

        '''
        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt
        x = self.array()
        d = x.shape[1]
        fig = plt.figure()
        if d == 1:
            plt.plot(x, *args, **kwargs)
        elif d == 2:
            plt.plot(x[:, 0], x[:, 1], *args, **kwargs)
        elif d == 3:
            ax = fig.add_subplot(111, projection='3d')
            ax.scatter(x[:, 0], x[:, 1], x[:, 2], *args, **kwargs)
        else:
            raise ValueError('The dimension must be less or equal than 3.')

    @abstractmethod
    def array(self):
        '''
        Return an array of shape (n, d) where n is the number of grid points
        and d is the dimension.

        Returns
        -------
        numpy.ndarray
            The TensorGrid as an array.

        '''

    @property
    @abstractmethod
    def size(self):
        '''
        Return the number of grid points.

        Returns
        -------
        int
            The number of grid points.

        '''


class FullTensorGrid(TensorGrid):
    '''
    Class FullTensorGrid: tensor product grid.

    '''

    def __init__(self, grids, dim=None):
        '''
        Constructor for the class FullTensorGrid.

        Parameters
        ----------
        grids : list or numpy.ndarray
            List containing arrays of shape (n_i, d_i), 1 <= i <= dim, or
            unidimensional list or numpy.ndarray replicated dim times
            (isotropic grid).
        dim : int, optional
            The dimension of the tensor grid. The default is None, indicating
            to infer it from grids.

        Returns
        -------
        None.

        '''

        tensap.TensorGrid.__init__(self)

        if dim is None:
            dim = len(grids)
        else:
            assert np.ndim(grids) == 1, \
                'The first argument must be a unidimensional grid.'
            grids = [np.reshape(grids, (-1, 1))]*dim
        grids = [np.atleast_2d(x) for x in grids]

        self.dim = dim
        self.shape = np.array([x.shape[0] for x in grids])
        self.grids = grids

    def array(self, ind=None):
        if ind is None:
            ind = self.multi_indices().array
        x = []
        for i in range(self.dim):
            x.append(self.grids[i][ind[:, i], :])
        return np.column_stack(x)

    def eval_at_indices(self, ind):
        return self.array(ind)

    @property
    def size(self):
        return np.prod(self.shape)

    def multi_indices(self):
        '''
        Return a set of multi-indices for indexing the grid.

        Returns
        -------
        tensap.MultiIndices
            Aset of multi-indices for indexing the grid.

        '''
        return tensap.MultiIndices.bounded_by(self.shape-1)

    def plot(self, y, *args):
        '''
        Plot the grid.

        Parameters
        ----------
        y : list or numpy.ndarray
            The values of the function on the grid.
        *args : tuple
            Parameters of the plot, see matplotlib.pyplot.plot for more
            information.

        Returns
        -------
        None.

        '''
        import matplotlib.pyplot as plt
        d = self.ndim()
        plt.figure()
        if d == 1:
            plt.plot(self.grids[0], y, *args)
        elif d == 2:
            from mpl_toolkits.mplot3d import Axes3D
            ax = plt.axes(projection='3d')
            x = self.array()
            ax.scatter(x[:, 0], x[:, 1], y, *args)
        else:
            raise ValueError('The dimension must be less or equal than 2.')

    @staticmethod
    def random(X, n):
        '''
        Generate a random FullTensorGrid from a RandomVector.

        Parameters
        ----------
        X : tensap.RandomVector
            The RandomVector used to generate the FullTensorGrid.
        n : list or numpy.ndarray
            Array of size d (or an integer for isotropic grid) containing
            the sizes of the grids in each dimension.

        Returns
        -------
        tensap.FullTensorGrid
            The random FullTensorGrid.

        '''
        d = X.size

        n = np.atleast_1d(n)
        if n.size == 1:
            n = np.tile(n, d)

        G = []
        for k in range(d):
            G.append(X.random_variables[k].random(n[k]))
        return FullTensorGrid(G)


class SparseTensorGrid(TensorGrid):
    '''
    Class SparseTensorGrid: sparse tensor product grid.

    '''

    def __init__(self, grids, indices, dim=None):
        '''
        Constructor for the class SparseTensorGrid

        Parameters
        ----------
        grids : list or numpy.ndarray
            List containing arrays of shape (n_i, d_i), 1 <= i <= dim, or
            unidimensional list or numpy.ndarray replicated dim times
            (isotropic grid).
        indices : tensap.MultiIndices
            The indices locating the non-zero coefficients (indices start at
            0).
        dim : int, optional
            The dimension of the tensor grid. The default is None, indicating
            to infer it from grids.

        Returns
        -------
        None.

        '''
        tensap.TensorGrid.__init__(self)

        T = tensap.FullTensorGrid(grids, dim)
        self.dim = T.dim
        self.grids = T.grids
        self.shape = T.shape
        self.indices = indices

    @property
    def size(self):
        return self.indices.cardinal()

    def array(self):
        ind = self.indices.to_list()

        x = list(self.grids)
        for k in range(len(x)):
            x[k] = x[k][ind[k], :]
        return np.hstack(x)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on learning in tree-based tensor format.

'''

from time import time
import numpy as np
import tensap

# %% Function to approximate
CHOICE = 2
if CHOICE == 1:
    ORDER = 5
    X = tensap.RandomVector(tensap.NormalRandomVariable(), ORDER)

    def fun(x):
        return 1 / (10 + x[:, 0] + 0.5*x[:, 1])**2
elif CHOICE == 2:
    fun, X = tensap.multivariate_functions_benchmark('borehole')
    ORDER = X.size
else:
    raise ValueError('Bad function choice.')

# %% Approximation basis
DEGREE = 8
ORTHONORMAL_BASES = True
if ORTHONORMAL_BASES:
    BASES = [tensap.PolynomialFunctionalBasis(
        x.orthonormal_polynomials(), range(DEGREE+1)) for
             x in X.random_variables]
else:
    BASES = [tensap.PolynomialFunctionalBasis(
        tensap.CanonicalPolynomials(), range(DEGREE+1)) for
             x in X.random_variables]
BASES = tensap.FunctionalBases(BASES)

# %% Training and test samples
NUM_TRAIN = 100
X_TRAIN = X.random(NUM_TRAIN)
Y_TRAIN = fun(X_TRAIN)

NUM_TEST = 10000
X_TEST = X.random(NUM_TEST)
Y_TEST = fun(X_TEST)

# %% Tree-based tensor format
# Tensor format
# 1 - Random tree and active nodes
# 2 - Tensor-Train
# 3 - Hierarchial Tensor-Train
# 4 - Binary tree
CHOICE = 3
if CHOICE == 1:
    print('Random tree with active nodes')
    ARITY = [2, 4]
    TREE = tensap.DimensionTree.random(ORDER, ARITY)
    IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
    SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE,
                                            tensap.SquareLossFunction())
elif CHOICE == 2:
    print('Tensor-train format')
    SOLVER = tensap.TreeBasedTensorLearning.tensor_train(
        ORDER, tensap.SquareLossFunction())
elif CHOICE == 3:
    print('Tensor Train Tucker')
    SOLVER = tensap.TreeBasedTensorLearning.tensor_train_tucker(
        ORDER, tensap.SquareLossFunction())
elif CHOICE == 4:
    print('Binary tree')
    TREE = tensap.DimensionTree.balanced(ORDER)
    IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
    SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE,
                                            tensap.SquareLossFunction())
else:
    raise NotImplementedError('Not implemented.')

# %% Random shuffling of the dimensions associated to the leaves
RANDOMIZE = True
if RANDOMIZE:
    SOLVER.tree.dim2ind = np.random.permutation(SOLVER.tree.dim2ind)
    SOLVER.tree = SOLVER.tree.update_dims_from_leaves()

# %% Learning in tree-based tensor format
SOLVER.bases = BASES
SOLVER.bases_eval = BASES.eval(X_TRAIN)
SOLVER.training_data = [None, Y_TRAIN]

SOLVER.tolerance['on_stagnation'] = 1e-6
SOLVER.tolerance['on_error'] = 1e-6

SOLVER.initialization_type = 'canonical'

SOLVER.linear_model_learning.regularization = False
SOLVER.linear_model_learning.basis_adaptation = True
SOLVER.linear_model_learning.error_estimation = True

SOLVER.test_error = True
SOLVER.test_data = [X_TEST, Y_TEST]
# SOLVER.bases_eval_test = BASES.eval(X_TEST)

SOLVER.rank_adaptation = True
SOLVER.rank_adaptation_options['max_iterations'] = 20
SOLVER.rank_adaptation_options['theta'] = 0.8
SOLVER.rank_adaptation_options['early_stopping'] = True
SOLVER.rank_adaptation_options['early_stopping_factor'] = 10

SOLVER.tree_adaptation = True
SOLVER.tree_adaptation_options['max_iterations'] = 1e2
# SOLVER.tree_adaptation_options['force_rank_adaptation'] = True

SOLVER.alternating_minimization_parameters['stagnation'] = 1e-10
SOLVER.alternating_minimization_parameters['max_iterations'] = 50

SOLVER.display = True
SOLVER.alternating_minimization_parameters['display'] = False

SOLVER.model_selection = True
SOLVER.model_selection_options['type'] = 'test_error'

# DMRG (inner rank adaptation)
DMRG = False
if DMRG:
    # Rank of the initialization
    SOLVER.rank = 3
    # Type of rank adaptation: can be 'dmrg' to perform classical DMRG (using
    # a truncation of the factor to adapt the rank) or 'dmrg_low_rank' to learn
    # the factor using a rank-adaptive learning
    # algorithm
    SOLVER.rank_adaptation_options['type'] = 'dmrg_low_rank'
    # Maximum alpha-rank for all the nodes of the tree
    SOLVER.rank_adaptation_options['max_rank'] = 10
    # If True: perform an alternating minimization with fixed tree and ranks
    # after the rank adaptation using DMRG
    SOLVER.rank_adaptation_options['post_alternating_minimization'] = False
    # Model selection type when type is 'dmrg_low_rank': can be 'cv_error' to
    # use a cross-validation estimator of the error or 'test_error' to use
    # the error on a test sample
    SOLVER.rank_adaptation_options['model_selection_type'] = 'cv_error'

T0 = time()
F, OUTPUT = SOLVER.solve()
T1 = time()
print(T1-T0)

# %% Displays
TEST_ERROR = SOLVER.loss_function.test_error(F, [X_TEST, Y_TEST])
print('Ranks: %s' % F.tensor.ranks)
print('Loo error = %2.5e' % OUTPUT['error'])
print('Test error = %2.5e' % TEST_ERROR)

F.tensor.plot(title='Active nodes')
F.tensor.tree.plot_dims(title='Dimensions associated to the leaf nodes')
F.tensor.tree.plot_with_labels_at_nodes(F.tensor.representation_rank,
                                        title='Representation ranks')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on tensor completion with tree-based tensor format.

'''

from enum import Enum
from time import time
import numpy as np
import tensap

np.random.seed(0)

# %% Generation of the tensor to recover
sz = [5, 8, 4, 3, 6]  # Size of the tensor
ORDER = len(sz)  # Order of the tensor
T = tensap.DimensionTree.linear(ORDER)
ranks = [1, 3, 3, 4, 3, 6, 3, 2, 4]
np.random.seed(0)
FTBT = tensap.TreeBasedTensor.randn(T, ranks, sz)
F = FTBT.full()

# %% Training and test samples
p = 0.3  # Proportion of known entries of the tensor

N = np.prod(sz)
NUM_TRAIN = np.int64(np.round(p*N))
loc_TRAIN = np.random.choice(N,NUM_TRAIN,replace=False)  # Random selection of n entries
indices_TRAIN = tensap.MultiIndices.ind2sub(sz, loc_TRAIN)
Y_TRAIN = F.eval_at_indices(indices_TRAIN.array)

NUM_TEST = N
loc_TEST = range(N)
indices_TEST = tensap.MultiIndices.ind2sub(sz, loc_TEST)
Y_TEST = F.eval_at_indices(indices_TEST.array)
print('Tensor to recover with ',p*100,'% of its entries known (',NUM_TRAIN,' entries):')
print(FTBT)

# %% Features creation: matrices containing non zero entries if the relative entry is known
FEATURES_TRAIN = []
for i in range(ORDER):
    M = np.zeros((NUM_TRAIN,sz[i]))
    np.put_along_axis(M, indices_TRAIN.array[:, i:i+1], 1.0, axis=1)
    FEATURES_TRAIN.append(M)


FEATURES_TEST = []
for i in range(ORDER):
    M = np.zeros((NUM_TEST,sz[i]))
    np.put_along_axis(M, indices_TEST.array[:, i:i+1], 1.0, axis=1)
    FEATURES_TEST.append(M)


# %% Tree-based tensor format
# Tensor format

class TensorFormat(str, Enum):
    RANDON_TREE_AND_ACTIVE_NODES = 'Random tree and active nodes'
    TENSOR_TRAIN = 'Tensor-Train'
    HIERARCHICAL_TENSOR_TRAIN = 'Hierarchial Tensor-Train'
    BINARY_TREE = 'Binary tree'


CHOICE = TensorFormat.HIERARCHICAL_TENSOR_TRAIN
if CHOICE == TensorFormat.RANDON_TREE_AND_ACTIVE_NODES:
    print('Random tree with active nodes')
    ARITY = [2, 4]
    TREE = tensap.DimensionTree.random(ORDER, ARITY)
    IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
    SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE,
                                            tensap.SquareLossFunction())
elif CHOICE == TensorFormat.TENSOR_TRAIN:
    print('Tensor-train format')
    SOLVER = tensap.TreeBasedTensorLearning.tensor_train(
        ORDER, tensap.SquareLossFunction())
elif CHOICE == TensorFormat.HIERARCHICAL_TENSOR_TRAIN:
    print('Tensor Train Tucker')
    SOLVER = tensap.TreeBasedTensorLearning.tensor_train_tucker(
        ORDER, tensap.SquareLossFunction())
elif CHOICE == TensorFormat.BINARY_TREE:
    print('Binary tree')
    TREE = tensap.DimensionTree.balanced(ORDER)
    IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
    SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE,
                                            tensap.SquareLossFunction())
else:
    raise NotImplementedError('Not implemented.')

TREE=SOLVER.tree
IS_ACTIVE_NODE=SOLVER.is_active_node

# %% Random shuffling of the dimensions associated to the leaves
RANDOMIZE = False
if RANDOMIZE:
    SOLVER.tree.dim2ind = np.random.permutation(SOLVER.tree.dim2ind)
    SOLVER.tree = SOLVER.tree.update_dims_from_leaves()

# %% Initial guess: known entries in a rank-1 tree-based tensor
guess = np.zeros(np.prod(sz))
guess[loc_TRAIN]=Y_TRAIN
guess = guess.reshape(sz,order='F')
guess = tensap.FullTensor(guess,order=ORDER,shape=sz)
tr = tensap.Truncator(tolerance=0, max_rank = 1)
guess = tr.hsvd(guess, SOLVER.tree,SOLVER.is_active_node)


# %% Learning in tree-based tensor format
SOLVER.bases_eval = FEATURES_TRAIN
SOLVER.training_data = [None, Y_TRAIN]

SOLVER.tolerance['on_stagnation'] = 1e-8
SOLVER.tolerance['on_error'] = 1e-8

SOLVER.initialization_type = 'canonical'

SOLVER.linear_model_learning.regularization = False
SOLVER.linear_model_learning.basis_adaptation = True
SOLVER.linear_model_learning.error_estimation = True

SOLVER.test_error = True
SOLVER.test_data = [None, Y_TEST]
SOLVER.bases_eval_test = FEATURES_TEST

SOLVER.rank_adaptation = True
SOLVER.rank_adaptation_options['max_iterations'] = 20
SOLVER.rank_adaptation_options['theta'] = 0.8
SOLVER.rank_adaptation_options['early_stopping'] = False
SOLVER.rank_adaptation_options['early_stopping_factor'] = 10

SOLVER.tree_adaptation = False
SOLVER.tree_adaptation_options['max_iterations'] = 1e2
# SOLVER.tree_adaptation_options['force_rank_adaptation'] = True

SOLVER.alternating_minimization_parameters['stagnation'] = 1e-10
SOLVER.alternating_minimization_parameters['max_iterations'] = 50

SOLVER.display = True
SOLVER.alternating_minimization_parameters['display'] = False

SOLVER.model_selection = True
SOLVER.model_selection_options['type'] = 'test_error'

# DMRG (inner rank adaptation)
DMRG = False
if DMRG:
    # Rank of the initialization
    SOLVER.rank = 3
    # Type of rank adaptation: can be 'dmrg' to perform classical DMRG (using
    # a truncation of the factor to adapt the rank) or 'dmrg_low_rank' to learn
    # the factor using a rank-adaptive learning
    # algorithm
    SOLVER.rank_adaptation_options['type'] = 'dmrg_low_rank'
    # Maximum alpha-rank for all the nodes of the tree
    SOLVER.rank_adaptation_options['max_rank'] = 10
    # If True: perform an alternating minimization with fixed tree and ranks
    # after the rank adaptation using DMRG
    SOLVER.rank_adaptation_options['post_alternating_minimization'] = False
    # Model selection type when type is 'dmrg_low_rank': can be 'cv_error' to
    # use a cross-validation estimator of the error or 'test_error' to use
    # the error on a test sample
    SOLVER.rank_adaptation_options['model_selection_type'] = 'cv_error'

T0 = time()
Fpred, OUTPUT = SOLVER.solve()
T1 = time()
print(T1-T0)

# ESTIMATED TENSOR
Fpred = Fpred.tensor

# %% Displays
TEST_ERROR = SOLVER.loss_function.test_error(Fpred.eval_at_indices(indices_TEST.array), [None, Y_TEST])
print('Ranks: %s' % Fpred.ranks)
print('Loo error = %2.5e' % OUTPUT['error'])
print('Test error = %2.5e' % TEST_ERROR)

Fpred.plot(title='Active nodes')
Fpred.tree.plot_dims(title='Dimensions associated to the leaf nodes')
Fpred.tree.plot_with_labels_at_nodes(Fpred.representation_rank,
                                        title='Representation ranks')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module lebesgue_variable.

'''

from numpy import array
from numpy.random import rand
import tensap
import numpy as np


class LebesgueMeasure(tensap.Measure):
    '''
    Class LebesgueMeasure.

    Attributes
    ----------
    a : float, optional
        The lower bound of the support of the random variable. The default
        is 0.
    b : float, optional
        The upper bound of the support of the random variable. The default
        is 1.

    '''

    def __init__(self, a=0, b=1):
        '''
        Lebesgue Measure on [a,b].

        Parameters
        ----------
    a : float, optional
        The lower bound of the support of the random variable. The default
        is 0.
    b : float, optional
        The upper bound of the support of the random variable. The default
        is 1.

        Returns
        -------
        None.

        '''
        tensap.Measure.__init__(self)

        self.a = a
        self.b = b

    def __repr__(self):
        return ('<{} on [{}, {}]>').format(self.__class__.__name__,
                                           self.a,
                                           self.b)

    def shift(self, m,s):
        '''
        If self a Lebesgue measure on [a,b], returns the Lebesgue measure on interval [m+sa,m+sb]

        Parameters
        ----------
        m : float
        s : float
        
        Returns
        -------
        tensap.LebesgueMeasure

        '''
        shifted = tensap.LebesgueMeasure(self.a,self.b)
        shifted.a = shifted.a*s+m
        shifted.b = shifted.b*s+m
        return shifted


    @staticmethod
    def ndim():
        '''
        Return the dimension of the random variable, equal to 1.

        Returns
        -------
        int
            The dimension of the random variable.

        '''
        return 1
    
    def support(self):
        '''
        Return the support of the measure.

        Returns
        -------
        numpy.ndarray

        '''
        return array([self.a, self.b])

    def __eq__(self, L2):
        if not (isinstance(self,  LebesgueMeasure) and
                isinstance(L2, LebesgueMeasure)):
            is_equal = False
        else:
            is_equal = True
            param_1 = self.get_parameters()
            param_2 = L2.get_parameters()
            for ind in zip(param_1, param_2):
                is_equal = is_equal and (ind[0] == ind[1])
        return is_equal
        
    def __neq__(self, L2):
        return not (self == L2)
        
    def mass(self):
        '''
        Return the mass of the measure.

        Returns
        -------
        numpy.ndarray

        '''
        
        return self.b - self.a
    
    def gauss_integration_rule(self, nb_pts):
        '''
        Return the nb_pts-points gauss integration rule associated with the
        measure of self, using Golub-Welsch algorithm.

        Parameters
        ----------
        nb_pts : int
            The number of integration points.

        Returns
        -------
        tensap.IntegrationRule
            The integration rule associated with the measure of self.

        '''
        RV = tensap.UniformRandomVariable(self.a,self.b)
        I = RV.gauss_integration_rule(nb_pts)
        I.weights = I.weights*self.mass()
        
        return I
    
    
    def orthonormal_polynomials(self, *max_degree):
        '''
        Return the max_degree-1 first orthonormal polynomials associated with
        the LebesgueMeasure on [a,b].

        Parameters
        ----------
        max_degree : int, optional
            The maximum degree of the returned polynomials. The default is
            None, choosing the default maximum degree associated with the
            constructor of the polynomials.

        Returns
        -------
        poly : tensap.OrthonormalPolynomials
            The generated orthonormal polynomials.

        '''
        poly = tensap.LegendrePolynomials(*max_degree)
        poly._orthogonal_polynomials_norms = poly._orthogonal_polynomials_norms*np.sqrt(self.mass())
        
        if self != LebesgueMeasure(-1, 1):
            poly = tensap.ShiftedOrthonormalPolynomials(poly,
                                                        (self.a+self.b)/2,
                                                        (self.b-self.a)/2)
        return poly

    def get_parameters(self):
        '''
        Return the parameters of the Lebesgue measure on [a,b].

        Returns
        -------
        float
            The lower bound of the support of the measure.
        float
            The upper bound of the support of the measure

        '''
        return self.a, self.b

    def random(self, n=1):
        return rand(int(n)) * (self.b - self.a) + self.a



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module loss_functions.

'''

from abc import ABC, abstractmethod
import numpy as np


class LossFunction(ABC):
    '''
    Class LossFunction.

    '''

    def risk_estimation(self, fun, sample, *args, nargout=1):
        '''
        Compute an estimation of the risk associated with the function fun and
        the loss function using the provided sample.

        Parameters
        ----------
        fun : function
            The function to which the risk is associated.
        sample : list
            The sample used to estimate the risk.
        *args : tuple
            Additional parameters used by a specific loss function.
        nargout : int, optional
            Specifies the number of output variables. The default is 1,
            returning only the risk. Set to 2 to also return the so-called
            estimated reference risk.

        Returns
        -------
        np.float
            The estimated risk.
        np.float
            The estimated reference risk.

        '''
        loss, loss_ref = self.eval(fun, sample, *args, nargout=2)
        risk = np.mean(loss)
        if nargout == 1:
            return risk
        risk_ref = np.mean(loss_ref)
        return risk, risk_ref

    @abstractmethod
    def eval(self, fun, sample, *args, nargout):
        '''
        Evaluate the loss function using the function fun and the provided
        sample.

        Parameters
        ----------
         fun : function
            The function used to evaluate the loss function
        sample : list
            The sample used to evaluate the loss function
        *args : tuple
            Additional parameters used by a specific loss function.
        nargout : int, optional
            Specifies the number of output variables. The default is 1,
            returning only the evaluations of the loss. Set to 2 to also return
            the so-called reference evaluations of the loss.

        Returns
        -------
        None.

        '''

    @abstractmethod
    def relative_test_error(self, fun, sample, *args):
        '''
        Compute the relative test error associated with the function fun and
        the loss function using the provided sample.

        Parameters
        ----------
        fun : function
            The function to which the error is associated.
        sample : list
            The sample used to estimate the risk.
        *args : tuple
            Additional parameters used by a specific loss function.

        Returns
        -------
        np.float
            The relative test error.

        '''

    @abstractmethod
    def test_error(self, fun, sample, *args):
        '''
        Compute the test error associated with the function fun and
        the loss function using the provided sample.

        Parameters
        ----------
        fun : function
            The function to which the error is associated.
        sample : list
            The sample used to estimate the risk.
        *args : tuple
            Additional parameters used by a specific loss function.

        Returns
        -------
        np.float
            The relative test error.

        '''


class SquareLossFunction(LossFunction):
    '''
    Class SquareLossFunction.

    Attributes
    ----------
    error_type : string, optional
        The error type. The default is 'relative'. Can also be 'absolute'.

    '''

    def __init__(self):
        '''
        Constructor for the class SquareLossFunction.

        Returns
        -------
        None.

        '''
        self.error_type = 'relative'

    def eval(self, fun, sample, *args, nargout=1):
        try:
            y_pred = fun.eval(sample[0])
        except Exception:
            try:
                y_pred = fun(sample[0])
            except Exception:
                y_pred = fun

        y_pred = np.atleast_2d(y_pred)
        y_true = np.atleast_2d(sample[1])
        loss = np.sum((y_pred - y_true)**2, 1)

        if nargout == 1:
            return loss
        loss_ref = np.sum(y_true**2, 1)
        return loss, loss_ref

    def relative_test_error(self, fun, sample, *args):
        risk, risk_ref = self.risk_estimation(fun, sample, *args, nargout=2)
        return np.sqrt(risk / risk_ref)

    def test_error(self, fun, sample, *args):
        if self.error_type == 'absolute':
            error = np.sqrt(self.risk_estimation(fun, sample, *args))
        elif self.error_type == 'relative':
            error = self.relative_test_error(fun, sample, *args)
        else:
            raise ValueError('The error_type property must be set to "risk" ' +
                             'or "relative".')
        return error


class DensityL2LossFunction(LossFunction):
    '''
    Class DensityL2LossFunction.

    Attributes
    ----------
    error_type : string, optional
        The error type. The default is 'relative'.

    '''

    def __init__(self):
        '''
        Constructor for the class DensityL2LossFunction.

        Returns
        -------
        None.

        '''
        self.error_type = 'risk'

    def eval(self, fun, sample, *args, nargout=1):
        try:
            y_pred = fun.eval(sample[0])
        except Exception:
            try:
                y_pred = fun(sample[0])
            except Exception:
                y_pred = fun

        try:
            l_ref = fun.norm()**2
        except Exception:
            try:
                l_ref = args[0]**2
            except Exception:
                print('Input fun must have a method norm, or its norm ' +
                      'must be provided in last argument.')
                l_ref = np.nan

        loss = l_ref - 2*y_pred
        if nargout == 1:
            return loss
        return loss, l_ref

    def relative_test_error(self, fun, sample, *args):
        raise NotImplementedError('Relative test error not available for ' +
                                  'DensityL2LossFunction.')

    def test_error(self, fun, sample, *args):
        assert self.error_type == 'risk', \
            'The error_type attribute must be set to "risk".'
        return self.risk_estimation(fun, sample, *args)


class CustomLossFunction(LossFunction):
    '''
    Class CustomLossFunction.

    Attributes
    ----------
    loss_function : tensorflow.function
        The custom loss function, defined using tensorflow operations.
    error_function : function, optional
        The function used to compute the test error. The default is None,
        indicating to use the risk function.
    relative_error_function : function, optional
        The function used to compute the relative test error. The default is
        None, indicating to use the risk function.
    error_type : string, optional
        The error type. The default is 'relative'. Can also be 'absolute'.

    '''

    def __init__(self, loss_function):
        self.loss_function = loss_function
        self.error_function = None
        self.relative_error_function = None
        self.error_type = 'absolute'

    def risk_estimation(self, fun, sample, *args, nargout=1):
        try:
            from tensorflow import reduce_mean
        except ImportError:
            reduce_mean = np.mean
        loss, loss_ref = self.eval(fun, sample, *args, nargout=2)
        risk = reduce_mean(loss)
        if nargout == 1:
            return risk
        risk_ref = reduce_mean(loss_ref)
        return risk, risk_ref

    def eval(self, fun, sample, *args, nargout=1):
        try:
            y_pred = fun.eval(sample[0])
        except Exception:
            try:
                y_pred = fun(sample[0])
            except Exception:
                y_pred = fun

        try:
            y_true = sample[1]
        except Exception:
            y_true = None

        loss = self.loss_function(y_true, y_pred)

        if nargout == 1:
            return loss
        loss_ref = self.loss_function(y_true, 0*y_pred)
        return loss, loss_ref

    def relative_test_error(self, fun, sample, *args):
        assert self.relative_error_function is not None, \
            ('Relative test error not available, please provide '
             'self.relative_error_function.')
        return self.relative_error_function(fun, sample)

    def test_error(self, fun, sample, *args):
        if self.error_function is None:
            error = self.risk_estimation(fun, sample, *args)
        else:
            error = self.error_function(fun, sample, *args)
        return error



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module empirical_random_variable.

'''

import scipy
import numpy as np
import tensap


class EmpiricalRandomVariable(tensap.RandomVariable):
    '''
    Class EmpiricalRandomVariable. A random variable fitted using gaussian
    kernel smoothing, this class gives best results in the case of normal
    distributions.

    Attributes
    ----------
    sample : numpy.array
        The sample used to generate the random variable.
    bandwidth : float
        The computed bandwidth for the kernel density estimator.

    '''

    def __init__(self, sample):
        '''
        Constructor for the class EmpiricalRandomVariable.

        A sample must be provided, which is used to fit a probability density
        function using Scott's rule.

        Parameters
        ----------
        sample : numpy.ndarray or list
            The sample used used to fit the probability density function using
            Scott's rule.

        Returns
        -------
        None.

        '''
        tensap.RandomVariable.__init__(self)

        self.sample = sample
        #  Scott's rule [Scott, D.W. (1992) Multivariate Density Estimation.
        # Theory, Practice and Visualization. New York: Wiley.]
        self.bandwidth = 3.5 * np.std(sample) * np.size(sample) ** (-1/3)

    def shift(self, b, s):
        '''
        Shift the random variable using the provided bias and scaling factor.

        Parameters
        ----------
        bias : float
            The bias.
        scaling : float
            The scaling factor.

        Returns
        -------
        tensap.EmpiricalRandomVariable
            The shifted random variable.

        '''
        return EmpiricalRandomVariable(s*self.sample+b)

    def transfer(self, Y, x):
        assert isinstance(Y, tensap.RandomVariable), \
            'The first argument must be a RandomVariable.'

        # If Y is the standard random variable associated to X
        if isinstance(Y, (EmpiricalRandomVariable,
                      tensap.EmpiricalRandomVariable)) and \
            np.linalg.norm(Y.sample - (self.sample - np.mean(self.sample)) /
                           np.std(self.sample)) / np.linalg.norm(Y.sample) < \
                np.finfo(float).eps:
            y = (x - np.mean(self.sample)) / np.std(self.sample)
        else:
            y = Y.icdf(self.cdf(x))
        return y

    def get_standard_random_variable(self):
        '''
        Return the standard empirical random variable with zero mean and
        unit standard deviation.

        Returns
        -------
        tensap.EmpiricalRandomVariable
            The standard empirical random variable.

        '''
        x = np.reshape(self.sample, [-1, 1])
        x = (x - np.tile(np.mean(x, 0), (x.shape[0], 1))) / \
            np.tile(np.std(x, 0), (x.shape[0], 1))
        return EmpiricalRandomVariable(np.ravel(x))

    def support(self):
        return np.array([-np.inf, np.inf])

    def orthonormal_polynomials(self, *args):
        p = tensap.EmpiricalPolynomials(self, *args)
        m = np.mean(self.sample)
        s = np.std(self.sample)
        if np.abs(m) > np.finfo(float).eps or \
                np.abs(s-1) > np.finfo(float).eps:
            p = tensap.ShiftedOrthonormalPolynomials(p, m, s)
        return p

    def get_parameters(self):
        return self.sample, self.bandwidth

    def cdf(self, x):
        kde = scipy.stats.gaussian_kde(self.sample, self.bandwidth)
        return np.vectorize(lambda z: kde.integrate_box_1d(-np.inf, z))(x)

    def icdf(self, x, *args, **kwargs):
        '''
        Compute the inverse cumulative distribution function (icdf) of the
        RandomVariable at points x.

        Parameters
        ----------
        x : float or list or numpy.ndarray
            The points at which the icdf is to be evaluated.
        *args, **kwargs : tuples
            Additional parameters for scipy.optimize's function brentq.

        Returns
        -------
        numpy.ndarray
            The evaluations of the icdf at points x.

        '''
        # TODO Optimize the method
        sup = self.truncated_support()
        return np.vectorize(lambda p: scipy.optimize.brentq(
            lambda y: self.cdf(y)-p, sup[0], sup[1], *args, **kwargs))(x)

    def pdf(self, x):
        xi, h = self.get_parameters()
        n = xi.size

        X = np.tile(np.reshape(x, [1, -1]), (n, 1))
        XI = np.tile(np.reshape(xi, [-1, 1]), (1, np.size(x)))
        return np.ravel(1/(n*h*np.sqrt(2*np.pi)) *
                        np.sum(np.exp(-0.5*(X-XI)**2/h**2), 0))

    def random_variable_statistics(self):
        return np.mean(self.sample), np.var(self.sample)

    def random(self, n=1):
        return self.icdf(np.random.rand(int(n)))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module tensor_principal_component_analysis.

'''

from copy import deepcopy
import numpy as np
import tensap


class TensorPrincipalComponentAnalysis:
    '''
    Class TensorPrincipalComponentAnalysis: principal component analysis of an
    algebraic tensor.

    Attributes
    ----------
    display : bool
        Boolean specifying the verbosity of the methods.
    pca_sampling_factor : int
        A factor to determine the number of samples N for the estimation of
        the principal components (1 by default):

        - if prescribed precision, N = pca_sampling_factor*N_alpha,
        - if prescribed rank, N = pca_sampling_factor*t.

    pca_adaptive_sampling : bool
        Adaptive sampling to determine the principal components with prescribed
        precision.
    tol : int or float or list or numpy.ndarray
        An array containing the prescribed relative precisions (the length
        depends on the format).
        If len(tol)==1, use the same value for all alpha.
        Set tol = inf to prescribe the rank.
    max_rank : int or list or numpy.ndarray
        An array containing the maximum alpha-ranks (the length depends on the
        format).
        If len(max_rank)==1, use the same value for all alpha.
        Set max_rank = np.inf to prescribe the precision.

    '''

    def __init__(self):
        '''
        Constructor for the class TensorPrincipalComponentAnalysis.

        Returns
        -------
        None.

        '''
        self.display = True
        self.pca_sampling_factor = 1
        self.pca_adaptive_sampling = False
        self.tol = 1e-8
        self.max_rank = np.inf

    def alpha_principal_components(self, fun, shape, alpha, tol, B_alpha,
                                   I_alpha):
        '''
        Evaluate the alpha-principal components of a tensor f.

        For alpha in {0,...,d-1}, it evaluates the alpha-principal components
        of a tensor f, meaning the principal components of the matricisations
        f_alpha(i_alpha,i_notalpha), where i_alpha and i_notalpha are groups of
        indices.

        It evaluates f_alpha on the product of a set of indices in dimension
        alpha (of size Nalpha) and a set of random indices (N samples) in the
        complementary dimensions.
        Then, it computes approximations of the alpha-principal components
        in a given basis phi_1(i_alpha) ... phi_Nalpha(i_alpha).

        If t is an integer, t is the rank (number of principal components).
        If t<1, the rank (number of principal components) is determined such
        that the relative error after truncation is t.

        Parameters
        ----------
        fun : fun or tensap.Function
            Function of d variables i_1, ..., i_d which returns the entries of
            the tensor.
        shape : list or numpy.ndarray
            The shape of the tensor.
        alpha : int
            An array containing a tuple in {0,...,d-1}.
        tol : int or float
            The number of principal components or a positive number smaller
            than 1 (tolerance).
        B_alpha : numpy.ndarray
            Array of shape (N_\alpha, N_\alpha) whose i-th column is the
            evaluation of phi_i at the set of indices i_alpha in Ialpha.
        I_alpha : numpy.ndarray
            Array of shape (N_alpha, #alpha) containing N_alpha tuples i_alpha.

        Returns
        -------
        pc : numpy.ndarray
            The principal components of the tensor.
        output : dict
            A dictionnary of outputs, containing the singular values
            corresponding to the principal components, as well as the set of
            indices at which the tensor has been evaluated.

        '''
        alpha = np.atleast_1d(alpha)
        B_alpha = np.atleast_2d(B_alpha)
        I_alpha = np.array(I_alpha)

        X = tensap.random_multi_indices(shape)
        d = len(shape)
        not_alpha = np.setdiff1d(range(d), alpha)

        if tol < 1:
            N = self.pca_sampling_factor * B_alpha.shape[1]
        else:
            N = self.pca_sampling_factor * tol
        N = int(np.ceil(N))

        X_not_alpha = tensap.RandomVector(X.random_variables[not_alpha])
        I_not_alpha = X_not_alpha.random(N)

        alpha_not_alpha = np.concatenate((alpha, not_alpha))
        ind = [np.nonzero(alpha_not_alpha == x)[0][0] for x in range(d)]

        output = {}
        if tol < 1 and self.pca_adaptive_sampling:
            A = tensap.FullTensor(np.zeros((I_alpha.shape[0], 0)), 2,
                                  [I_alpha.shape[0], 0])
            for k in range(N):
                product_grid = tensap.FullTensorGrid(
                    [I_alpha, I_not_alpha[k, :]]).array()
                A_k = np.linalg.solve(B_alpha, fun(product_grid[:, ind]))
                A.data = np.column_stack((A.data, A_k))
                pc, sin_val = A.principal_components(tol)
                if sin_val[-1, -1] < 1e-15 or pc.shape[1] < \
                        np.ceil((k+1) / self.pca_sampling_factor):
                    break
            output['number_of_evaluations'] = I_alpha.shape[0] * (k+1)
        else:
            product_grid = tensap.FullTensorGrid([I_alpha,
                                                  I_not_alpha]).array()
            A = fun(product_grid[:, ind])
            A = np.reshape(A, [B_alpha.shape[0], N], 'F')
            A = np.linalg.solve(B_alpha, A)
            A = tensap.FullTensor(A, 2, [B_alpha.shape[0], N])
            pc, sin_val = A.principal_components(tol)
            output['number_of_evaluations'] = I_alpha.shape[0] * N

        output['singular_values'] = sin_val
        output['samples'] = product_grid
        return pc, output

    def hopca(self, fun, shape):
        '''
        Return the set of alpha-principal components of an algebraic tensor,
        for all alpha in {0,1,...,d-1}.

        For prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d).

        For prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d).

        See also the documentation of the class
        TensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : fun or tensap.Function
            Function of d variables i_1, ..., i_d which returns the entries of
            the tensor.
        shape : list or numpy.ndarray
            The shape of the tensor.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        f_pc : list
            List of the alpha-principal components of the tensor.
        output : list
            List containing the outputs of the method
            alpha_principal_components.

        '''
        solver = deepcopy(self)
        d = len(shape)

        if np.ndim(self.tol) == 0 or len(self.tol) == 1:
            solver.tol = np.full(d, self.tol)
        elif len(self.tol) != d:
            raise ValueError('tol should be a scalar or an array of length d.')

        if np.ndim(self.max_rank) == 0 or len(self.max_rank) == 1:
            solver.max_rank = np.full(d, self.max_rank)
        elif len(self.max_rank) != d:
            raise ValueError('max_rank should be a scalar or an array of ' +
                             'length d.')

        f_pc = []
        output = []
        for alpha in range(d):
            I_alpha = np.reshape(np.arange(shape[alpha]), (-1, 1))
            B_alpha = np.eye(shape[alpha])
            tol_alpha = np.min((solver.tol[alpha], solver.max_rank[alpha]))
            f_pc_alpha, output_alpha = \
                solver.alpha_principal_components(fun, shape, alpha, tol_alpha,
                                                  B_alpha, I_alpha)
            f_pc.append(f_pc_alpha)
            output.append(output_alpha)
        return f_pc, output

    def tucker_approximation(self, fun, shape):
        '''
        Approximation of a tensor of order d in Tucker format based on a
        Principal Component Analysis.

        For a prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d).

        For a prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d).

        See also the documentation of the class
        TensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : fun or tensap.Function
            Function of d variables i_1, ..., i_d which returns the entries of
            the tensor.
        shape : list or numpy.ndarray
            The shape of the tensor.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        tensap.TreeBasedTensor
            A tensor in tree based format with a trivial tree.
        dict
            Dictionnary containing the outputs of the method.

        '''
        solver = deepcopy(self)
        d = len(shape)
        tree = tensap.DimensionTree.trivial(d)

        if np.ndim(self.tol) == 1 and len(self.tol) == d:
            tol = solver.tol
            solver.tol = np.zeros(d+1)
            solver.tol[tree.dim2ind-1] = tol
        elif np.ndim(self.tol) == 1 and len(self.tol) > 1:
            raise ValueError('tol should be a scalar or an array of length d.')

        if np.ndim(self.max_rank) == 1 and len(self.max_rank) == d:
            rank = solver.max_rank
            solver.max_rank = np.zeros(d+1)
            solver.max_rank[tree.dim2ind-1] = rank
        elif np.ndim(self.max_rank) == 1 and len(self.max_rank) > 1:
            raise ValueError('max_rank should be a scalar or an array of ' +
                             'length d.')

        return solver.tree_based_approximation(fun, shape, tree)

    def tt_approximation(self, fun, shape):
        '''
        Approximation of a tensor of order d in tensor train format based on a
        Principal Component Analysis.

        For a prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d-1).

        For a prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d-1).

        See also the documentation of the class
        TensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : fun or tensap.Function
            Function of d variables i_1, ..., i_d which returns the entries of
            the tensor.
        shape : list or numpy.ndarray
            The shape of the tensor.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        tensap.TreeBasedTensor
            A tensor in tree based format with a linear tree.
        dict
            Dictionnary containing the outputs of the method.

        '''
        solver = deepcopy(self)
        d = len(shape)
        tree = tensap.DimensionTree.linear(d)
        is_active_node = np.full(tree.nb_nodes, True)
        is_active_node[tree.dim2ind[1:]-1] = False
        rep_tt = np.nonzero(is_active_node)[0]
        rep_tt = np.flip(rep_tt[1:])

        if np.ndim(self.tol) == 1 and len(self.tol) == d-1:
            tol = solver.tol
            solver.tol = np.zeros(tree.nb_nodes)
            solver.tol[rep_tt] = tol
        elif np.ndim(self.tol) == 1 and len(self.tol) > 1:
            raise ValueError('tol should be a scalar or an array of length ' +
                             'd-1.')

        if np.ndim(self.max_rank) == 1 and len(self.max_rank) == d-1:
            rank = solver.max_rank
            solver.max_rank = np.zeros(tree.nb_nodes)
            solver.max_rank[rep_tt] = rank
        elif np.ndim(self.max_rank) == 1 and len(self.max_rank) > 1:
            raise ValueError('max_rank should be a scalar or an array of ' +
                             'length d-1.')

        return solver.tree_based_approximation(fun, shape, tree,
                                               is_active_node)

    def tree_based_approximation(self, fun, shape, tree, is_active_node=None):
        '''
        Approximation of a tensor of order d in tree based tensor format based
        on a Principal Component Analysis.

        For a prescribed precision, set TPCA.max_rank = np.inf and TPCA.tol to
        the desired precision (possibly an array of length d-1).

        For a prescribed rank, set TPCA.tol = np.inf and TPCA.max_rank to the
        desired rank (possibly an array of length d-1).

        See also the documentation of the class
        TensorPrincipalComponentAnalysis.

        Parameters
        ----------
        fun : fun or tensap.Function
            Function of d variables i_1, ..., i_d which returns the entries of
            the tensor.
        shape : list or numpy.ndarray
            The shape of the tensor.
        tree : tensap.DimensionTree
            The required dimension tree.
        is_active_node : list or numpy.ndarray, optional
            An array of booleans indicating which nodes of the tree are active.
            The default is None, settings all the nodes active.

        Raises
        ------
        ValueError
            If the provided tolerance and max ranks are not correct.

        Returns
        -------
        tensap.TreeBasedTensor
            A tensor in tree based format.
        dict
            Dictionnary containing the outputs of the method.

        '''
        solver = deepcopy(self)
        d = len(shape)

        if is_active_node is None:
            is_active_node = np.full(tree.nb_nodes, True)

        if (np.ndim(self.tol) == 0 or len(self.tol) == 1) and self.tol < 1:
            solver.tol /= np.sqrt(np.count_nonzero(is_active_node)-1)

        if np.ndim(self.tol) == 0 or len(self.tol) == 1:
            solver.tol = np.full(tree.nb_nodes, solver.tol)
        elif len(self.tol) != tree.nb_nodes:
            raise ValueError('tol should be a scalar or an array of length ' +
                             'tree.nb_nodes.')

        if np.ndim(self.max_rank) == 0 or len(self.max_rank) == 1:
            solver.max_rank = np.full(tree.nb_nodes, self.max_rank)
        elif len(self.max_rank) != tree.nb_nodes:
            raise ValueError('max_rank should be a scalar or an array of ' +
                             'length tree.nb_nodes.')

        grids = [np.reshape(np.arange(x), (-1, 1)) for x in shape]
        alpha_basis = np.empty(tree.nb_nodes, dtype=object)
        alpha_grids = np.empty(tree.nb_nodes, dtype=object)
        outputs = np.empty(tree.nb_nodes, dtype=object)
        samples = np.empty(tree.nb_nodes, dtype=object)
        tensors = [[]]*tree.nb_nodes
        number_of_evaluations = 0
        for nu in range(d):
            alpha = tree.dim2ind[nu]
            B_alpha = np.eye(shape[nu])
            if is_active_node[alpha-1]:
                tol_alpha = np.min((solver.tol[alpha-1],
                                    solver.max_rank[alpha-1]))
                pc_alpha, outputs[alpha-1] = \
                    solver.alpha_principal_components(fun, shape, nu,
                                                      tol_alpha, B_alpha,
                                                      grids[nu])
                samples[alpha-1] = outputs[alpha-1]['samples']
                shape_alpha = [shape[nu], pc_alpha.shape[1]]
                tensors[alpha-1] = tensap.FullTensor(pc_alpha, 2, shape_alpha)

                B_alpha = np.matmul(B_alpha, pc_alpha)
                I_alpha = tensap.magic_indices(B_alpha)[0]
                alpha_grids[alpha-1] = grids[nu][I_alpha, :]
                alpha_basis[alpha-1] = B_alpha[I_alpha, :]

                number_of_evaluations += outputs[alpha-1][
                    'number_of_evaluations']
                if solver.display:
                    print('alpha = %i : rank = %i, nb_eval = %i' %
                          (alpha, shape_alpha[-1],
                           outputs[alpha-1]['number_of_evaluations']))
            else:
                alpha_grids[alpha-1] = grids[nu]
                alpha_basis[alpha-1] = B_alpha

        for level in np.arange(np.max(tree.level), 0, -1):
            for alpha in np.intersect1d(tree.nodes_with_level(level),
                                        tree.internal_nodes):
                S_alpha = tree.children(alpha)
                B_alpha = TensorPrincipalComponentAnalysis.\
                    _tensor_product_b_alpha(alpha_basis[S_alpha-1])
                alpha_grids[alpha-1] = \
                    tensap.FullTensorGrid(alpha_grids[S_alpha-1]).array()

                tol_alpha = np.min((solver.tol[alpha-1],
                                    solver.max_rank[alpha-1]))
                pc_alpha, outputs[alpha-1] = \
                    solver.alpha_principal_components(fun, shape,
                                                      tree.dims[alpha-1],
                                                      tol_alpha,
                                                      B_alpha,
                                                      alpha_grids[alpha-1])
                samples[alpha-1] = outputs[alpha-1]['samples']
                shape_alpha = np.concatenate(([x.shape[1] for
                                              x in alpha_basis[S_alpha-1]],
                                             [pc_alpha.shape[1]]))
                tensors[alpha-1] = tensap.FullTensor(pc_alpha, len(S_alpha)+1,
                                                     shape_alpha)

                B_alpha = np.matmul(B_alpha, pc_alpha)
                I_alpha = tensap.magic_indices(B_alpha)[0]
                alpha_grids[alpha-1] = alpha_grids[alpha-1][I_alpha, :]
                alpha_basis[alpha-1] = B_alpha[I_alpha, :]
                number_of_evaluations += outputs[alpha-1][
                    'number_of_evaluations']
                if solver.display:
                    print('alpha = %i : rank = %i, nb_eval = %i' %
                          (alpha, shape_alpha[-1],
                           outputs[alpha-1]['number_of_evaluations']))

        alpha = tree.root
        S_alpha = tree.children(alpha)
        B_alpha = TensorPrincipalComponentAnalysis.\
            _tensor_product_b_alpha(alpha_basis[S_alpha-1])
        I_alpha = tensap.FullTensorGrid(alpha_grids[S_alpha-1]).array()
        shape_alpha = [x.shape[1] for x in alpha_basis[S_alpha-1]]
        ind = [np.nonzero(tree.dims[alpha-1] == x)[0][0] for x in range(d)]
        tensors[alpha-1] = tensap.FullTensor(
            np.linalg.solve(B_alpha, fun(I_alpha[:, ind])), len(S_alpha),
            shape_alpha)
        alpha_grids[alpha-1] = I_alpha
        number_of_evaluations += I_alpha.shape[0]
        samples[alpha-1] = I_alpha
        if solver.display:
            print('Interpolation - nb_eval = %i' % I_alpha.shape[0])

        f = tensap.TreeBasedTensor(tensors, tree)

        output = {'number_of_evaluations': number_of_evaluations,
                  'samples': samples,
                  'alpha_basis': alpha_basis,
                  'alpha_grids': alpha_grids,
                  'outputs': outputs}

        return f, output

    @staticmethod
    def _tensor_product_b_alpha(Bs):
        '''
        Function used in the method tree_based_approximation.

        Parameters
        ----------
        Bs : list
            List containing s matricices B1 , ..., Bs where Bi is a
            n[i]-by-r[i] array.

        Returns
        -------
        numpy.ndarray
            An array of shape prod(n)-by-prod(r) whose entries are
            B[I , J] = B1[i_1, j_1] ... Bs[i_s, j_s]
            with I = (i_1, ..., is) and J = (j1, ..., js).

        '''
        Bs = [tensap.FullTensor(x, 2, x.shape) for x in Bs]
        B = Bs[0]
        for k in np.arange(1, len(Bs)):
            B = B.tensordot(Bs[k], 0)
        return B.matricize(np.arange(0, B.order-1, 2)).data



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module function.

'''

from abc import abstractmethod
import numpy as np
import tensap


class Function:
    '''
    Class Function.

    Attributes
    ----------
    dim : int
        The dimension of the input of the function.
    measure : tensap.Measure
        The measure associated with the function.
    output_shape : int or list or numpy.ndarray
        The shape of the output of the function.
    evaluation_at_multiple_points : bool
        Indicates if the function can be evaluated at multiple points at once.
    store : bool
        Indicates if the Function object should store the evaluations of the
        function.

    '''

    def __init__(self):
        '''
        Constructor for the class Function.

        Returns
        -------
        None.

        '''
        self.dim = []
        self.measure = None
        self.output_shape = 1
        self.evaluation_at_multiple_points = True
        self.store = False
        self._x_stored = []
        self._y_stored = []

    def __call__(self, x, return_f=False):
        if np.ndim(x) == 1:
            x = np.expand_dims(x, 1)

        if self.store:
            y = self.store_eval(x)
            if np.ndim(y) == 2 and np.shape(y)[1] == 1:
                y = np.squeeze(y, axis=1)
            return y
        else:
            y = self.eval(x)
            if np.ndim(y) == 2 and np.shape(y)[1] == 1:
                y = np.squeeze(y, axis=1)
            f = self
            if return_f:
                return y, f
            return y

    def store_eval(self, x):
        '''
        Evaluate the function, reuising previous evaluations if possible, and
        storing the new evaluations in self.

        Parameters
        ----------
        x : numpy.ndarray
            The input points.

        Returns
        -------
        y : numpy.ndarray
            The evaluations of the Function.
        tensap.Function
            The Function with stored evaluations, for future reuse.

        '''
        if np.ndim(x) < 2:
            x = np.reshape(x, [np.size(x), 1])

        if self.store and np.size(self._y_stored) != 0:
            ind_2, ind_1 = np.nonzero(np.all(x == self._x_stored[:,
                                                                 np.newaxis],
                                             axis=2))

            if np.prod(self.output_shape) != 1:
                self._y_stored = np.reshape(self._y_stored,
                                            (self._y_stored.shape[0], -1),
                                            order='F')

            y = np.zeros((x.shape[0], self._y_stored.shape[1]))
            y[ind_1, :] = self._y_stored[ind_2, :]

            x_new = x[np.setdiff1d(range(x.shape[0]), ind_1), :]
            if x_new.size != 0:
                y_new = self.eval(x_new)
                y[np.setdiff1d(range(x.shape[0]), ind_1), :] = \
                    np.reshape(y_new, (y_new.shape[0], -1))
                self._x_stored = np.vstack((self._x_stored, x_new))
                self._y_stored = np.vstack((self._y_stored,
                                            np.reshape(y_new,
                                                       (y_new.shape[0], -1))))
            y = np.reshape(y, np.concatenate(([y.shape[0]],
                                              np.atleast_1d(
                                                  self.output_shape))),
                           order='F')
            if np.prod(self.output_shape) != 1:
                self._y_stored = np.reshape(self._y_stored, np.concatenate(([
                    self._y_stored.shape[0]], np.atleast_1d(
                        self.output_shape))), order='F')
        else:
            y = self.eval(x)
            if np.ndim(y) < 2:
                y = np.reshape(y, [np.size(y), 1])

            self.store = True
            self._x_stored = x
            self._y_stored = y

        return np.squeeze(y), self

    def fplot(self, support=None, n_points=100, *args, **kwargs):
        '''
        Plot the function on a support using a given number of points.

        Parameters
        ----------
        support : list or numpy.ndarray, optional
            The support of the plot. The default is None, indicating to use
            the truncated_support of self.measure.
        n_points : int, optional
            The number of points used for the plot. The default is 100.
        *args : tuple
            Additional parameters used by the function matplotlib.pyplot.plot.

        Returns
        -------
        None.

        '''
        import matplotlib.pyplot as plt
        if support is None:
            support = self.measure.truncated_support()

        x = np.linspace(support[0], support[1], int(n_points))
        plt.plot(x, self.eval(x), *args, **kwargs)
        plt.show()

    def surf(self, n=None, *args):
        '''
        Surface plot of the bivariate function.

        Parameters
        ----------
        n : list or numpy.ndarray, optional
            The number of points used for the surface plot in each dimension.
            The default is [1000, 1000].
        *args : tuple
            Additional parameters used by matplotlib.pyplot's plot_surface
            function.

        Returns
        -------
        ax : matplotlib.axes._subplots.AxesSubplot
            The surface plot as a matplotlib.axes._subplots.AxesSubplot object.

        '''
        assert self.measure is not None, 'Attribute measure is empty.'
        assert self.dim == 2, \
            ('The function should be a bivariate function, use the partial ' +
             'evaluation for higher-dimensional function.')

        from mpl_toolkits.mplot3d import Axes3D
        import matplotlib.pyplot as plt

        if n is None:
            n = [1000, 1000]

        sup = self.measure.truncated_support()

        if np.size(n) == 1:
            n = np.tile(n, 2)

        grids = [np.linspace(sup[0][0], sup[0][1], n[0]),
                 np.linspace(sup[1][0], sup[1][1], n[1])]
        grids[0] = grids[0][1:-1]
        grids[1] = grids[1][1:-1]
        grids = [np.reshape(x, (x.size, -1)) for x in grids]

        grid = tensap.FullTensorGrid(grids)
        fg = self.eval_on_tensor_grid(grid)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        x, y = np.meshgrid(grids[0], grids[1])
        ax.plot_surface(x, y, fg.data, *args)
        plt.show()

        return ax

    def partial_evaluation(self, not_alpha, x_not_alpha):
        '''
        Return the partial evaluation of a function
        f(x) = f(x_alpha,x_not_alpha), a function
        f_alpha(.) = f(., x_not_alpha) for fixed values x_not_alpha of the
        variables with indices not_alpha.

        Parameters
        ----------
        not_alpha : list or numpy.ndarray
            The indices of the fixed variables.
        x_not_alpha : numpy.ndarray
            The points at which the function is evaluated in the dimensions
            not_alpha.

        Raises
        ------
        ValueError
            If the Function has an empty attribute dim.

        Returns
        -------
        f_alpha : tensap.UserDefinedFunction
            The partial evaluation of the Function.

        '''
        if self.dim is None or np.size(self.dim) == 0:
            raise ValueError('The Function has an empty attribute dim.')

        alpha = np.setdiff1d(range(self.dim), not_alpha)
        ind = [np.nonzero(np.concatenate((alpha, not_alpha)) == x)[0][0] for
               x in range(self.dim)]

        def fun(x_alpha):
            grid = tensap.FullTensorGrid([x_alpha, x_not_alpha])
            return self.eval(grid.array()[:, ind])

        f_alpha = tensap.UserDefinedFunction(fun, alpha.size)
        f_alpha.store = self.store
        f_alpha.evaluation_at_multiple_points = \
            self.evaluation_at_multiple_points
        f_alpha.measure = self.measure.marginal(alpha)

        return f_alpha

    def random(self, n=1, measure=None):
        '''
        Evaluates the function at n points x drawn randomly according to the
        ProbabilityMeasure in measure if provided, or in self.measure.

        Parameters
        ----------
        n : int, optional
            The number of random evaluations. The default is 1.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure according to which the points x are drawn.
            The default is None, indicating to use self.measure.

        Raises
        ------
        ValueError
            If the provided measure is not a tensap.ProbabilityMeasure.

        Returns
        -------
        numpy.ndarray
            The evaluations of the function at the points x.
        x : numpy.ndarray
            The points at which the function is to be evaluated.

        '''
        if measure is None:
            if isinstance(self.measure, tensap.ProbabilityMeasure):
                measure = self.measure
            else:
                raise ValueError('Must provide a ProbabilityMeasure.')
        elif not isinstance(self.measure, tensap.ProbabilityMeasure):
            raise ValueError('Must provide a ProbabilityMeasure.')

        x = measure.random(n)
        return self.eval(x), x

    def eval_on_tensor_grid(self, x):
        '''
        Evaluate the Function on a grid x.

        Parameters
        ----------
        x : tensap.TensorGrid
            The tensap.ensorGrid used for the evaluation..

        Raises
        ------
        NotImplementedError
            If the method is not implemented.
        ValueError
            If x is not a tensap.TensorGrid object.

        Returns
        -------
        fx : numpy.ndarray
            The evaluation of the Function on the grid.

        '''
        x_a = x.array()
        fx = self.eval(x_a)

        if isinstance(x, tensap.SparseTensorGrid):
            if np.all(self.output_shape == 1):
                fx = tensap.SparseTensor(fx, x.indices, x.shape)
            else:
                raise NotImplementedError('Method not implemented.')
        elif isinstance(x, tensap.FullTensorGrid):
            if np.all(self.output_shape == 1):
                shape = x.shape
                if self.dim > 1:
                    fx = np.reshape(fx, shape, order='F')
            else:
                shape = np.concatenate((np.atleast_1d(x.shape),
                                        np.atleast_1d(self.output_shape)))
                fx = np.reshape(fx, shape, order='F')
            fx = tensap.FullTensor(fx, np.size(shape), shape)
        else:
            raise ValueError('A TensorGrid object must be provided.')

        return fx

    def test_error(self, g, n=1000, measure=None):
        '''
        Compute the test error associated with the function, using a function g
        or some of its evaluations as a reference. A measure can be provided
        to randomly draw the test input data.

        Parameters
        ----------
        g : tensap.Function or numpy.ndarray
            The reference function or evaluations of it.
        n : int or numpy.ndarray, optional
            The test sample size, or the test input data. The default is 1000
            test input points.
        measure : tap.ProbabilityMeasure, optional
            A probability measure used to draw the test input data. The default
            is None, indicating to either use self.measure, g.measure or the
            provided input points.

        Returns
        -------
        err_l2 : float
            The L2 error.
        err_linf : float
            The L-infinity error.

        '''
        if measure is not None:
            x_test = measure.random(n)
            g_x_test = g.eval(x_test)
            err_l2, err_linf = self.test_error(g_x_test, x_test)
        else:
            if isinstance(g, tensap.Function) and np.size(n) == 1:
                if self.measure is not None:
                    measure = self.measure
                else:
                    measure = g.measure
                err_l2, err_linf = self.test_error(g, n, measure)
            elif np.size(n) != 1:
                x_test = n
                n = np.shape(x_test)[0]
                f_x_test = self.eval(x_test)
                if isinstance(g, tensap.Function):
                    g_x_test = g.eval(x_test)
                else:
                    assert np.shape(g)[0] == n, \
                        ('The number of evaluations does not match the ' +
                         'number of points.')
                    g_x_test = np.array(g)

                f_x_test = np.reshape(f_x_test, (n, -1), order='F')
                g_x_test = np.reshape(g_x_test, (n, -1), order='F')
                err_l2 = np.linalg.norm(f_x_test - g_x_test) / \
                    np.linalg.norm(g_x_test)
                err_linf = np.linalg.norm(np.sqrt(np.sum((f_x_test -
                                                          g_x_test)**2, 1)),
                                          np.inf) / \
                    np.linalg.norm(np.sqrt(np.sum(g_x_test**2, 1)), np.inf)

        return err_l2, err_linf

    @abstractmethod
    def eval(self, x):
        '''
        Evaluate the function at the points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the function is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the function at the points x.

        '''



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on learning in tree-based tensor format a tensorized function, with
an example of model selection using a slope heuristic instead of the error on a
test sample.

'''

from time import time
import numpy as np
import matplotlib.pyplot as plt
import tensap

# %% Function to approximate: identification of a function f(x) with a function
# g(i1,...,id,y) (see also tutorial_TensorizedFunction)
R = 4  # Resolution
B = 5  # Scaling factor
ORDER = R+1

X = tensap.UniformRandomVariable(0, 1)
Y = tensap.UniformRandomVariable(0, 1)

CHOICE = 1
if CHOICE == 1:
    def FUN(x):
        return np.sin(10*np.pi*(2*x+0.5))/(4*x+1) + (2*x-0.5)**4
elif CHOICE == 2:
    def FUN(x):
        return (np.sin(4*np.pi*x) + 0.2*np.cos(16*np.pi*x))**(x < 0.5) + \
            (2*x-1)*(x >= 0.5)
else:
    raise ValueError('Bad function choice.')

T = tensap.Tensorizer(B, R, 1, X, Y)
TENSORIZED_FUN = T.tensorize(FUN)
TENSORIZED_FUN.fun.evaluation_at_multiple_points = True

# %% Approximation basis
DEGREE = 5
H = tensap.PolynomialFunctionalBasis(Y.orthonormal_polynomials(),
                                     range(DEGREE+1))
BASES = T.tensorized_function_functional_bases(H)

# %% Training and test samples
NUM_TRAIN = 200
X_TRAIN = X.random(NUM_TRAIN)
Y_TRAIN = FUN(X_TRAIN)
X_TRAIN = T.map(X_TRAIN)  # Identification of X_TRAIN with (i_1,...,i_d,y)

NUM_TEST = 1000
X_TEST = X.random(NUM_TEST)
Y_TEST = FUN(X_TEST)
X_TEST = T.map(X_TEST)

# %% Tree-based tensor format
# Tensor format
# 1 - Random tree and active nodes
# 2 - Tensor-Train
# 3 - Hierarchial Tensor-Train
# 4 - Binary tree
CHOICE = 3
if CHOICE == 1:
    print('Random tree with active nodes')
    ARITY = [2, 4]
    TREE = tensap.DimensionTree.random(ORDER, ARITY)
    IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
    SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE,
                                            tensap.SquareLossFunction())
elif CHOICE == 2:
    print('Tensor-train format')
    SOLVER = tensap.TreeBasedTensorLearning.tensor_train(
        ORDER, tensap.SquareLossFunction())
elif CHOICE == 3:
    print('Tensor Train Tucker')
    SOLVER = tensap.TreeBasedTensorLearning.tensor_train_tucker(
        ORDER, tensap.SquareLossFunction())
elif CHOICE == 4:
    print('Binary tree')
    TREE = tensap.DimensionTree.balanced(ORDER)
    IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
    SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE,
                                            tensap.SquareLossFunction())
else:
    raise NotImplementedError('Not implemented.')

# %% Random shuffling of the dimensions associated to the leaves
RANDOMIZE = True
if RANDOMIZE:
    SOLVER.tree.dim2ind = np.random.permutation(SOLVER.tree.dim2ind)
    SOLVER.tree = SOLVER.tree.update_dims_from_leaves()

# %% Learning in tree-based tensor format
SOLVER.bases = BASES
SOLVER.bases_eval = BASES.eval(X_TRAIN)
SOLVER.training_data = [None, Y_TRAIN]

# Let the solver go to the maximum number of iterations, to perform model
# selection a posteriori using the test error and a slope heuristic
SOLVER.tolerance['on_stagnation'] = 0
SOLVER.tolerance['on_error'] = 0
SOLVER.rank_adaptation_options['early_stopping'] = False

SOLVER.initialization_type = 'canonical'

SOLVER.linear_model_learning.regularization = False
SOLVER.linear_model_learning.basis_adaptation = True
SOLVER.linear_model_learning.error_estimation = True

SOLVER.test_error = True
SOLVER.test_data = [X_TEST, Y_TEST]
# SOLVER.bases_eval_test = BASES.eval(X_TEST)

SOLVER.rank_adaptation = True
SOLVER.rank_adaptation_options['max_iterations'] = 50
SOLVER.rank_adaptation_options['theta'] = 0.8

SOLVER.tree_adaptation = True
SOLVER.tree_adaptation_options['max_iterations'] = 100
# SOLVER.tree_adaptation_options['force_rank_adaptation'] = True

SOLVER.alternating_minimization_parameters['stagnation'] = 1e-10
SOLVER.alternating_minimization_parameters['max_iterations'] = 50

SOLVER.display = True
SOLVER.alternating_minimization_parameters['display'] = False

SOLVER.model_selection = True
SOLVER.model_selection_options['type'] = 'test_error'

T0 = time()
F, OUTPUT = SOLVER.solve()
T1 = time()
print(T1-T0)

# %% Displays
TEST_ERROR = SOLVER.loss_function.test_error(F, [X_TEST, Y_TEST])
print('\nRanks: %s' % F.tensor.ranks)
print('Loo error = %2.5e' % OUTPUT['error'])
print('Test error = %2.5e' % TEST_ERROR)

F.tensor.plot(title='Active nodes')
F.tensor.tree.plot_dims(title='Dimensions associated to the leaf nodes')
F.tensor.tree.plot_with_labels_at_nodes(F.tensor.representation_rank,
                                        title='Representation ranks')

plt.figure()
x_lin = np.linspace(0, 1, 1000)
plt.plot(x_lin, FUN(x_lin), x_lin, F(T.map(x_lin)))
plt.legend(('True function', 'Approximation'))
plt.show()

# %% Model selection using a slope heuristic instead of the test error
SEL = tensap.ModelSelection()
# Complexity of each model: standard complexity based on the method storage
SEL.data['complexity'] = SEL.complexity(OUTPUT['iterates'],
                                        fun='storage',
                                        c_type='standard')
# Empirical risk associated with each model
SEL.data['empirical_risk'] = [SOLVER.loss_function.risk_estimation(
    x, (X_TRAIN, Y_TRAIN)) for x in OUTPUT['iterates']]
# Shape of the penalization function
SEL.pen_shape = lambda x: x / NUM_TRAIN
# Gap factor used in the slope heuristic
SEL.gap_factor = 2

LAMBDA_HAT, M_HAT, LAMBDA_PATH, M_PATH = SEL.slope_heuristic()

print('Model selected using a test sample:         %i, test error = %2.5e.' %
      (OUTPUT['selected_model_number'], OUTPUT['test_error']))
print('Model selected using the slope heuristic:   %i, test error = %2.5e.' %
      (M_HAT, OUTPUT['test_error_iterations'][M_HAT]))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module functional_tensor.

'''

from copy import deepcopy
import numpy as np
import tensap


class FunctionalTensor(tensap.Function):
    '''
    Class FunctionalTensor.

    Attributes
    ----------
    tensor : Tensor or tensap.FunctionalTensor, optional
        The tensor of the FunctionalTensor. The default is None.
    bases : list or tensap.FunctionalBases, optional
        The functional bases of the FunctionalTensor. The default is None.
    fdims : list or numpy.ndarray, optional
        The dimensions corresponding to the bases. The default is None.

    '''

    def __init__(self, tensor=None, bases=None, fdims=None):
        '''
        Constructor for the FunctionalTensor.

        Parameters
        ----------
        tensor : Tensor or tensap.FunctionalTensor, optional
            The tensor of the FunctionalTensor. The default is None.
        bases : list or tensap.FunctionalBases, optional
            The functional bases of the FunctionalTensor. The default is None.
        fdims : list or numpy.ndarray, optional
            The dimensions corresponding to the bases. The default is None.

        Raises
        ------
        ValueError
            If the provided objects are not of the expected types.

        Returns
        -------
        None.

        '''
        tensap.Function.__init__(self)

        if isinstance(tensor, FunctionalTensor):
            tensor = tensor.tensor

        self.tensor = tensor

        if bases is not None and \
                not isinstance(bases, (tensap.FunctionalBases, list)):
            raise ValueError('Must provide a FunctionalBases object, or a ' +
                             'cell of bases evaluations.')
        self.bases = bases
        if fdims is None:
            if tensor.order != len(bases):
                raise ValueError('Bases must contain as many bases as the ' +
                                 'order of the tensor, with possible empty ' +
                                 'elements.')
            fdims = range(tensor.order)
        else:
            if len(fdims) != len(bases):
                raise ValueError('The number of functional dimensions must ' +
                                 'correspond to the number of bases in bases.')
        self.fdims = np.array(fdims)
        if isinstance(bases, list):
            self.evaluated_bases = True
        else:
            self.evaluated_bases = False
            self.measure = self.bases.measure
            self.dim = np.sum(self.bases.ndim())

    def __plus__(self, y):
        if isinstance(y, FunctionalTensor) and self.bases == y.bases:
            tensor = self.tensor + y.tensor
        else:
            raise NotImplementedError('Method not implemented.')
        return FunctionalTensor(tensor, self.bases, self.fdims)

    def is_random(self):
        '''
        Determine if self is random.

        Returns
        -------
        bool
            Boolean indicating if if self is random.

        '''
        return isinstance(self.bases.measure, tensap.ProbabilityMeasure)

    def mean(self, *measure):
        '''
        Compute the expectation of the random variable self(measure) if
        measure is provided, and of self(self.bases.measure) otherwise.

        Parameters
        ----------
        *measure : tensap.RandomVector, optional
            The measure used for the computation of the mean. If not provided,
            indicates to use self.bases.measure.

        Returns
        -------
        float or Tensor
            The mean of the function.

        '''
        bases_eval = self.bases.mean(None, *measure)
        return self.tensor.tensor_vector_product(bases_eval,
                                                 self.fdims).tolist()

    def expectation(self, *measure):
        '''
        Compute the expectation of the random variable self(measure) if
        measure is provided, and of self(self.bases.measure) otherwise.

        Parameters
        ----------
        *measure : tensap.RandomVector, optional
            The measure used for the computation of the expectation. If not
            provided, indicates to use self.bases.measure.

        Returns
        -------
        float or Tensor
            The expectation of the function.

        '''
        return self.mean(*measure)

    def variance(self, *measure):
        '''
        Compute the variance of the random variable self(measure) if
        measure is provided, and of self(self.bases.measure) otherwise.

        Parameters
        ----------
        *measure : tensap.RandomVector, optional
            The measure used for the computation of the variance. If not
            provided, indicates to use self.bases.measure.

        Returns
        -------
        var : float or Tensor
            The variance of the function.

        '''
        mean = self.expectation(*measure)
        if np.isscalar(mean):
            var = self.dot_product_expectation(self, None, *measure) - mean**2
        else:
            raise NotImplementedError('Method not implemented.')
        return var

    def std(self, *measure):
        ''''
        Compute the standard deviation of the random variable self(measure) if
        measure is provided, and of self(self.bases.measure) otherwise.

        Parameters
        ----------
        *measure : tensap.RandomVector, optional
            The measure used for the computation of the standard deviation. If
            not provided, indicates to use self.bases.measure.

        Returns
        -------
        v : float or Tensor
            The standard deviation of the function.

        '''
        return np.sqrt(self.variance(*measure))

    def dot_product_expectation(self, f_2, fdims=None, *measure):
        '''
        Computes the expectation of self(X)f_2(X) where X is the random vector
        associated with self.bases if measure is not provided, and measure
        otherwise.

        For tensor-valued functions of X (len(X)<self.order), fdims specifies
        the dimensions of self and f_2 corresponding to theRandomVector X.

        Parameters
        ----------
        f_2 : tensap.FunctionalTensor
            The second functional tensor of the product.
        fdims : list of numpy.ndarray, optional
            Specifies the dimensions of self and f_2 corresponding to
            theRandomVector X. The default is None, indicating all the
            dimensions
        *measure : tensap.RandomVector, optional
            The measure used for the computation of the product. If not
            provided, indicates to use self.bases.measure.

        Raises
        ------
        ValueError
            If the two tensors do not have the same order and fdims is not
            specified.
        NotImplementedError
            If the bases of self and f_2 are not equal.

        Returns
        -------
        float or Tensor
            The expectation of self(X)f_2(X).

        '''
        if fdims is None:
            if self.tensor.order == f_2.tensor.order:
                fdims = range(self.tensor.order)
            else:
                raise ValueError('Tensors u and v do not have the same ' +
                                 'order, must specify fdims.')
        if self.bases == f_2.bases:
            gram_matrix = self.bases.gram_matrix(fdims, *measure)
        else:
            raise NotImplementedError('Method not implemented.')
        tensor = self.tensor.tensor_matrix_product(gram_matrix, fdims)
        return tensor.dot(f_2.tensor)

    def norm(self, *measure):
        '''
        Return the L^2 norm of self(X), with X = measure if provided, and
        X = self.bases.measure otherwise.

        If self.evaluatedBases is true, without additional information, return
        the canonical norm of self.tensor.

        Parameters
        ----------
        *measure : tensap.RandomVector, optional
            The measure used for the computation of the norm. If not provided,
            indicates to use self.bases.measure.

        Returns
        -------
        float or Tensor
            The L^2 norm of self(X).

        '''
        if not self.evaluated_bases:
            gram_matrix = self.bases.gram_matrix(range(self.tensor.order),
                                                 *measure)
        else:
            gram_matrix = [np.eye(x.shape[1]) for x in self.bases]
        tensor = self.tensor.tensor_matrix_product(gram_matrix,
                                                   range(self.tensor.order))
        return np.sqrt(tensor.dot(self.tensor))

    def conditional_expectation(self, dims, *args):
        '''
        Compute the conditional expectation of self with respect to the random
        variables dims (a subset of [1, ..., d]).

        The expectation with respect to other variables (in the complementary
        set of dims) is taken with respect to the probability measure given by
        a tensap.RandomVector if provided as an additional argument, or with
        respect to the probability measure associated with the corresponding
        bases of self.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The dimensions of the random variables with respect to which the
            conditional expectation is to be computed.
        *args : tuple
            Tuple containing a tensap.randomVector giving the probability
            measure of the variables other than the ones in dims. If not
            provided, the measure is infered from self.bases.measure.

        Returns
        -------
        tensap.FunctionalTensor
            The conditional expectation of self with respect to the random
            variables dims, as a len(dims)-order tensor.

        '''
        dims = np.atleast_1d(dims)
        if np.all([isinstance(x, bool) for x in dims]):
            dims = np.nonzero(dims)[0]

        d = self.tensor.order
        if np.size(dims) == 0:
            return self.expectation(*args)

        dims = np.sort(dims)
        assert np.size(self.fdims) == d and \
            np.array_equal(self.fdims, range(d)), \
            ('Method not implemented for self.fdims different from ' +
             'range(d).')

        dims_C = np.setdiff1d(range(len(self.bases)), dims)
        if dims_C.size == 0:
            return deepcopy(self)

        H = self.bases.mean(dims_C, *args)
        t = self.tensor.tensor_vector_product(H, dims_C)

        bases = self.bases.keep_bases(dims)
        # TODO Take into account the mapping when implemented

        out = FunctionalTensor(t, bases)
        if self.measure is not None:
            out.measure = self.measure.marginal(dims)

        return out

    def variance_conditional_expectation(self, alpha):
        '''
        Compute the variance of the conditional expectation of self in
        dimensions in alpha.

        Parameters
        ----------
        alpha : list or numpy.ndarray
            Array containing the dimensions (either explicitely or using
            booleans) in which the variance of the conditional expectation is
            computed.

        Returns
        -------
        v : numpy.ndarray
            The variance of the conditional expectation of self in
            dimensions in alpha.

        '''
        alpha = np.atleast_2d(alpha)
        m = self.expectation()
        v = np.zeros(alpha.shape[0])
        for i in range(alpha.shape[0]):
            u = alpha[i, :]
            if np.all([isinstance(x, bool) for x in u]):
                u = np.nonzero(u)[0]
            if u.size == 0:
                v[i] = 0
            else:
                mu = self.conditional_expectation(u)
                v[i] = mu.dot_product_expectation(mu) - m**2
        return v

    def eval(self, x, *dims):
        '''
        Evaluate self at the points x.

        If dims is provided, compute the partial evaluations of self at points
        x in dimensions in dims.

        Parameters
        ----------
        x : list or numpy.ndarray or None
            The points at which the function is to be evaluated. If x is None
            and self.evaluated_bases, evaluates the function using the
            evaluations of the bases.
        *dims : list or numpy.ndarray, optional
            The dimensions of the partial evaluation. If not provided,
            evaluate the function in all dimensions.

        Returns
        -------
        numpu.ndarray or Tensor
            The evaluations of self at the points x.

        '''
        if self.evaluated_bases:
            bases_eval = self.bases
        else:
            bases_eval = self.bases.eval(x, *dims)
        return self.eval_with_bases_evals(bases_eval, *dims)

    def __mul__(self, f_2):
        if isinstance(f_2, (FunctionalTensor, tensap.FunctionalTensor)):
            b = self.bases.kron(f_2.bases)
            t = self.tensor.kron(f_2.tensor)
            out = FunctionalTensor(t, b)
            if isinstance(out.tensor, tensap.TreeBasedTensor) and \
                    out.tensor.ranks[out.tensor.tree.root-1] > 1:
                if self.tensor.ranks[self.tensor.tree.root-1] != \
                        f_2.tensor.ranks[f_2.tensor.tree.root-1]:
                    raise ValueError('Wrong tensor shapes.')
                else:
                    root = out.tensor.tree.root
                    c = out.tensor.tensors[root-1]
                    n = self.tensor.ranks[root-1]
                    s = [':']*c.order
                    s[-1] = np.arange(n**2, step=n)
                    c = c.sub_tensor(*s)
                    out.tensor.tensors[root-1] = c
                    out.tensor.ranks[root-1] = n
        else:
            out = deepcopy(self)
            out.tensor = out.tensor * f_2
        return out

    def parameter_gradient_eval(self, alpha, x=None, *args):
        '''
        Compute the gradient of the function with respect to its alpha-th
        parameter, evaluated at some points.

        Parameters
        ----------
        alpha : int
            The number of the parameter with respect to which compute the
            gradient of self.
        x : list or numpy.ndarray, optional
            The points at which the gradient is to be evaluated. The default is
            None, indicating to use self.bases if self.evaluated_bases is True.

        Raises
        ------
        ValueError
            If no input points are provided.

        Returns
        -------
        grad : Tensor
            The gradient of the function with respect to its alpha-th
            parameter, evaluated at some points.

        '''
        if self.evaluated_bases:
            bases_eval = self.bases
        elif x is not None:
            bases_eval = self.bases.eval(x)
        else:
            raise ValueError('Must provide the evaluation points or the ' +
                             'bases evaluations.')

        dims = np.arange(self.tensor.order)
        if isinstance(self.tensor, tensap.TreeBasedTensor):
            # Compute fH, the TimesMatrixEvalDiag of f with bases_eval in all
            # the dimensions except the ones associated with alpha (if alpha
            # is a leaf node) or with the inactive children of alpha (if
            # alpha is an internal node). The tensor fH is used to compute
            # the gradient of f with respect to f.tensor.tensors[alpha-1].
            tree = self.tensor.tree
            if tree.is_leaf[alpha-1]:
                dims = dims[tree.dim2ind != alpha]
            else:
                children = tree.children(alpha)
                ind = tensap.fast_intersect(
                    tree.dim2ind,
                    children[np.logical_not(
                        self.tensor.is_active_node[children-1])])
                dims = dims[np.logical_not(np.isin(tree.dim2ind, ind))]

            if np.all(self.tensor.is_active_node):
                fH = self.tensor.tensor_matrix_product([bases_eval[x] for x
                                                        in dims], dims)
            else:
                remaining_dims = np.arange(self.tensor.order)
                tensors = np.array(self.tensor.tensors)
                dim2ind = np.array(tree.dim2ind)

                for leaf in tensap.fast_intersect(tree.dim2ind[dims],
                                                  self.tensor.active_nodes):
                    dims = tensap.fast_setdiff(
                        dims, np.nonzero(tree.dim2ind == leaf)[0][0])
                    tensors[leaf-1] = self.tensor.tensors[leaf-1].\
                        tensor_matrix_product(bases_eval[
                            np.nonzero(tree.dim2ind == leaf)[0][0]], 0)

                for pa in np.unique(tree.parent(tensap.fast_setdiff(
                        tree.dim2ind[dims], self.tensor.active_nodes))):
                    ind = tensap.fast_intersect(tree.dim2ind[dims],
                                                tree.children(pa))
                    ind = tensap.fast_setdiff(ind, self.tensor.active_nodes)
                    dims_loc = np.array([np.nonzero(x == tree.dim2ind)[0][0]
                                         for x in ind])
                    if len(ind) > 1:
                        tensors[pa-1] = self.tensor.tensors[pa-1].\
                            tensor_matrix_product_eval_diag([bases_eval[x] for
                                                             x in dims_loc],
                                                            tree.child_number(
                                                                ind)-1)
                        remaining_dims = tensap.fast_setdiff(remaining_dims,
                                                             dims_loc[1:])
                        if np.all(np.logical_not(self.tensor.is_active_node[
                                tree.children(pa)-1])):
                            dim2ind[dims_loc[0]] = tree.parent(
                                tree.dim2ind[dims_loc[0]])
                        else:
                            dims = tensap.fast_setdiff(dims, dims_loc[0])
                        dim2ind[dims_loc[1:]] = 0
                        perm = np.concatenate((
                            [tree.child_number(ind[0])-1],
                            tensap.fast_setdiff(np.arange(tensors[pa-1].order),
                                                tree.child_number(ind[0])-1)))
                        tensors[pa-1] = tensors[pa-1].itranspose(perm)
                    elif len(ind) == 1:
                        dims = dims[dims != dims_loc]
                        tensors[pa-1] = self.tensor.tensors[pa-1].\
                            tensor_matrix_product([bases_eval[x] for
                                                   x in dims_loc],
                                                  tree.child_number(ind)-1)
                        dim2ind[dims_loc] = tree.dim2ind[dims_loc]

                keep_ind = tensap.fast_setdiff(np.arange(tree.nb_nodes),
                                               tree.dim2ind[dims]-1)
                adj_mat = tree.adjacency_matrix[np.ix_(keep_ind, keep_ind)]
                dim2ind = dim2ind[dim2ind != 0]

                ind = np.zeros(tree.nb_nodes)
                ind[tensap.fast_setdiff(np.arange(tree.nb_nodes),
                                        keep_ind)] = 1
                ind = np.cumsum(ind).astype(int)
                dim2ind -= ind[dim2ind-1]
                alpha = alpha - ind[alpha-1]

                tree = tensap.DimensionTree(dim2ind, adj_mat)
                fH = tensap.TreeBasedTensor(tensors[keep_ind], tree)
                fH = fH.remove_unique_children()
                bases_eval = [bases_eval[x] for x in remaining_dims]
        else:
            if alpha <= self.tensor.order:
                dims = np.delete(dims, alpha-1)
            fH = self.tensor.tensor_matrix_product([bases_eval[x] for
                                                    x in dims], dims)

        grad = fH.parameter_gradient_eval_diag(alpha, bases_eval)
        if isinstance(self.tensor, tensap.TreeBasedTensor) and \
                not tree.is_leaf[alpha-1]:
            # If the order of the children has been modified in grad, compute
            # the inverse permutation.
            ch = tree.children(alpha)
            perm_1 = np.argsort(np.concatenate((
                np.atleast_1d(ch[fH.is_active_node[ch-1]]),
                np.atleast_1d(ch[np.logical_not(fH.is_active_node[ch-1])]))))

            if alpha == tree.root:
                perm_2 = []
            else:
                perm_2 = [fH.tensors[alpha-1].order]

            if alpha != tree.root and self.tensor.ranks[tree.root-1] > 1:
                perm_3 = [grad.order-1]
            else:
                perm_3 = []

            grad = grad.transpose(np.concatenate(([0], perm_1+1,
                                                  perm_2, perm_3)).astype(int))

        return grad

    def parameter_gradient_eval_dmrg(self, alpha, x=None, dmrg_type='dmrg',
                                     *args):
        if self.evaluated_bases:
            bases_eval = self.bases
        elif x is not None:
            bases_eval = self.bases.eval(x)
        else:
            raise ValueError('Must provide the evaluation points or the ' +
                             'bases evaluations.')

        dims = np.arange(self.tensor.order)
        if isinstance(self.tensor, tensap.TreeBasedTensor):
            # Compute fH, the TimesMatrixEvalDiag of f with bases_eval in all
            # the dimensions except the ones associated with alpha (if alpha
            # is a leaf node) or with the inactive children of alpha (if
            # alpha is an internal node). The tensor fH is used to compute
            # the gradient of f with respect to f.tensor.tensors[alpha-1].
            tree = self.tensor.tree
            if tree.is_leaf[alpha-1]:
                dims = dims[tree.dim2ind != alpha]
            else:
                children = tree.children(alpha)
                ind = tensap.fast_intersect(
                    tree.dim2ind,
                    children[np.logical_not(
                        self.tensor.is_active_node[children-1])])
                dims = dims[np.logical_not(np.isin(tree.dim2ind, ind))]

            fH = self.tensor.tensor_matrix_product([bases_eval[x] for x
                                                    in dims], dims)
        else:
            if alpha <= self.tensor.order:
                dims = np.delete(dims, alpha-1)
            fH = self.tensor.tensor_matrix_product([bases_eval[x] for
                                                    x in dims], dims)

        grad, g_alpha, g_gamma = \
            fH.parameter_gradient_eval_diag_dmrg(alpha, bases_eval)

        if isinstance(self.tensor, tensap.TreeBasedTensor):
            # If the order of the children has been modified in grad, compute
            # the inverse permutation.
            ch = tree.children(alpha)

            if ch.size == 0:
                perm_1 = np.array([0])
            else:
                perm_1 = np.argsort(np.concatenate((
                    np.atleast_1d(ch[fH.is_active_node[ch-1]]),
                    np.atleast_1d(
                        ch[np.logical_not(fH.is_active_node[ch-1])]))))
            gamma = tree.parent(alpha)
            ch = tensap.fast_setdiff(tree.children(gamma), alpha)
            perm_1b = np.argsort(np.concatenate((
                    np.atleast_1d(ch[fH.is_active_node[ch-1]]),
                    np.atleast_1d(
                        ch[np.logical_not(fH.is_active_node[ch-1])]))))

            if dmrg_type == 'dmrg':
                perm_1 = np.concatenate((perm_1, perm_1.size+perm_1b))
                perm_2 = []
                if alpha != tree.root and gamma != tree.root:
                    perm_2 = [fH.tensors[alpha-1].order +
                              fH.tensors[gamma-1].order-2]
                perm_3 = []
                if alpha != tree.root and self.tensor.ranks[tree.root-1] > 1:
                    perm_3 = [grad.order-1]
                grad = grad.transpose(np.concatenate(
                    ([0], perm_1+1, perm_2, perm_3)).astype(int))
            elif dmrg_type == 'dmrg_low_rank':
                g_alpha = g_alpha.transpose(np.concatenate(([0], perm_1+1)))
                perm_2 = []
                if gamma != tree.root:
                    # TODO Checks
                    perm_2 = [fH.tensors[gamma-1].order-1]
                perm_3 = []
                if alpha != tree.root and self.tensor.ranks[tree.root-1] > 1:
                    perm_3 = [grad.order-1]
                g_gamma = g_gamma.transpose(np.concatenate(
                    ([0], perm_1b+1, perm_2, perm_3)).astype(int))

                grad = [g_alpha, g_gamma]
            else:
                raise ValueError('Wrong DMRG type.')
        return grad

    def eval_derivative(self, n, x, *dims):
        '''
        Evaluate the n-th order derivative of self at the points x.

        If dims is provided, compute the partial evaluations of the n-th order
        derivative of self at points x n dimensions in dims.

        Parameters
        ----------
        n : int
            The order of derivation.
        x : list or numpy.ndarray
            The points at which the function is to be evaluated.
        *dims : list or numpy.ndarray, optional
            The dimensions of the partial evaluation. If not provided,
            evaluate the function in all dimensions.

        Returns
        -------
        numpu.ndarray or Tensor
            The evaluations of the n-th derivative of self at the points x.

        '''
        bases_eval = self.bases.eval_derivative(n, x, *dims)
        return self.eval_with_bases_evals(bases_eval, *dims)

    def derivative(self, n):
        '''
        Compute the n-th order derivative of self.

        Parameters
        ----------
        n : int
            The order of derivation.

        Returns
        -------
        df : tensap.FunctionalTensor
            The n-th order derivative of self.

        '''
        df = deepcopy(self)
        df.bases = self.bases.derivative(n)
        return df

    def eval_on_grid(self, x, dims=None):
        '''
        Compute evaluations of self at points x.

        Parameters
        ----------
        x : list
            List such that x[k] contains the grid associated with the (k+1)-th
            variable.
        dims : list or numpy.ndarray, optional
            Array indicating the dimensions associated with x. The default is
            None, indicating all the dimensions.

        Returns
        -------
        out : FunctionalTensor or tensap.Tensor
            The evaluations of self at points x.

        '''
        if dims is None:
            dims = np.arange(self.bases.length())
        H = self.bases.eval(x, dims)

        if np.size(dims) == self.tensor.order:
            out = self.tensor.tensor_matrix_product(H, dims)
        else:
            out = deepcopy(self)
            out.tensor = self.tensor.tensor_matrix_product(H, dims)
            out.bases = out.bases.remove_bases(dims)
            out.fdims = out.fdims[np.setdiff1d(
                np.range(np.size(out.fdims)), dims)]
            if np.size(out.fdims) == 0:
                out = out.tensor
        return out

    def random(self, *args, **kwargs):
        return self.random_dims(range(len(self.fdims)), *args, **kwargs)

    def random_dims(self, dims, *args, nargout=1):
        '''
        Evaluate the function in dimensions dims using n points drawn randomly
        according to measure if provided, or to
        self.bases.measure.marginal(dims) otherwise.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The dimensions of the bases to be evaluated.
        n : int, optional
            The number of random evaluations. The default is 1.
        measure : tensap.ProbabilityMeasure, optional
            The probability measure used for the generation of the input
            points. The default is None, indicating to use
            self.measure.marginal(dims).

        Returns
        -------
        bases_eval : list or numpy.ndarray
            Random evaluations of the function.
        x : numpy.ndarray
            The input points, grouped by basis.

        '''
        bases_eval, x = self.bases.random_dims(dims, *args, nargout=2)
        if nargout == 1:
            return self.eval_with_bases_evals(bases_eval, dims)
        return self.eval_with_bases_evals(bases_eval, dims), x

    def get_random_vector(self):
        '''
        Return the RandomVector associated with self.bases.

        Returns
        -------
        tensap.RandomVector
            The RandomVector associated with self.bases.

        '''
        return self.bases.get_random_vector()

    def eval_with_bases_evals(self, bases_eval, dims=None):
        '''
        Evaluate the function self, given evaluations of self.bases.

        Parameters
        ----------
        bases_eval : list or numpy.ndarray
            The evaluations of self.bases.
        dims : list or numpy.ndarray, optional
            The dimensions of the evaluation. The default is None, indicating
            all the dimensions.

        Returns
        -------
        out : numpy.ndarray or Tensor
            The (partially) evaluated function.

        '''
        if dims is None:
            dims = range(len(self.bases))
        if len(dims) == 1 and not isinstance(bases_eval, list):
            bases_eval = [bases_eval]
        if len(dims) == self.tensor.order:
            out = self.tensor.tensor_matrix_product_eval_diag(
                bases_eval).numpy()
        else:
            out = deepcopy(self)
            fdims_eval = out.fdims[dims]
            out.tensor = out.tensor.tensor_matrix_product_eval_diag(bases_eval,
                                                                    fdims_eval)
            fdims_eval.sort()
            old_dims = np.setdiff1d(range(self.tensor.order), fdims_eval[2:])
            if out.tensor.shape[fdims_eval[0]] == 1:
                out.tensor = out.tensor.squeeze(fdims_eval[0])
                old_dims = np.delete(old_dims, fdims_eval[0])
            out.bases = out.bases.remove_bases(dims)
            out.fdims = np.delete(out.fdims, dims)
            out.fdims = np.nonzero(np.isin(old_dims, out.fdims))[0]
            if out.fdims.size == 0:
                out = out.tensor
                if out.order == 1:
                    out = out.numpy()
        return out

    def storage(self):
        '''
        Return the storage requirement of the FunctionalTensor.

        Returns
        -------
        int
            The storage requirement of the FunctionalTensor.

        '''
        return self.tensor.storage()



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module sparse_tensor.

'''

from copy import deepcopy
import numpy as np
from scipy.sparse import lil_matrix, coo_matrix
import tensap


class SparseTensor:
    '''
    Class SparseTensor.

    Attributes
    ----------
    data : numpy.ndarray
        The values of the tensor at the entries in indices.
    indices : tensap.MultiIndices
        The set of multi-indices corresponding to the non-zero coefficients of
        the tensor.
    order : int
        The order of the tensor.
    shape : numpy.ndarray
        The shape of the tensor.

    '''

    def __init__(self, data=None, indices=None, shape=None):
        '''
        Constructor for the class SparseTensor.

        Parameters
        ----------
        data : list or numpy.ndarray, optional
            The values of the tensor at the entries in indices, or the tensor
            itself. The default is None.
        indices : tensap.MultiIndices, optional
            The set of multi-indices corresponding to the non-zero coefficients
            of the tensor. The default is None, indicating to infer it from
            the tensor provided in data.
        shape : list or numpy.ndarray, optional
            The shape of the tensor. The default is None, indicating to infer
            it from the tensor provided in data.

        Raises
        ------
        ValueError
            If the provided arguments are wrong.

        Returns
        -------
        None.

        '''
        if data is None and indices is None and shape is None:
            self.data = []
            self.order = 0
            self.indices = []
            self.shape = []
        elif data is not None and indices is None and shape is None:
            if isinstance(data, (list, np.ndarray)):
                self.order = np.ndim(data)
                self.shape = np.shape(data)

                rep = np.nonzero(data)
                self.data = data[rep]
                self.indices = tensap.MultiIndices(
                    np.hstack([np.reshape(x, [-1, 1]) for x in rep]))
            elif np.all([hasattr(data, x) for x in ['data', 'order',
                                                    'shape', 'indices']]):
                self.data = data.data
                self.order = data.order
                self.shape = data.shape
                self.indices = data.indices
            else:
                raise ValueError('Wrong input arguments.')
        elif data is not None and indices is not None and shape is not None:
            assert isinstance(indices, tensap.MultiIndices), \
                'Argument indices must be a MultiIndices.'

            self.indices = indices
            self.order = indices.ndim()
            self.shape = shape
            self.data = data
            assert np.size(data) == indices.cardinal(), \
                'data and indices must have the same number of elements.'
        else:
            raise ValueError('Wrong input arguments.')
        self.data = np.squeeze(self.data)
        self.shape = np.squeeze(self.shape)

    @property
    def size(self):
        '''
        Compute the size of the tensor. Equivalent to self.storage().

        Returns
        -------
        numpy.ndarray
            The size of the tensor.

        '''
        return np.prod(self.shape)

    def storage(self):
        '''
        Return the storage complexity of the SparseTensor.

        Returns
        -------
        int
            The storage complexity of the SparseTensor.

        '''
        return self.size

    def sparse_storage(self):
        '''
        Return the sparse storage complexity of the SparseTensor.

        Returns
        -------
        int
            The sparse storage complexity of the SparseTensor.

        '''
        return self.count_non_zero()

    def count_non_zero(self):
        '''
        Return the number of non-zero coefficients of the SparseTensor.
        Equivalent to self.sparse_storage().

        Returns
        -------
        int
            The number of non-zero coefficients of the SparseTensor.

        '''
        return self.indices.cardinal()

    @property
    def ndim(self):
        '''
        Compute the order of the tensor. Equivalent to self.order.

        Returns
        -------
        int
            The order of the tensor.

        '''
        return np.size(self.shape)

    def full(self):
        '''
        Convert the SparseTensor to a tensap.FullTensor.

        Returns
        -------
        y : tensap.FullTensor
            The SparseTensor as a tensap.FullTensor.

        '''
        y = tensap.FullTensor(np.zeros(self.shape))
        ind = tuple(self.indices.to_list())
        y.data[ind] = self.data
        return y

    def numpy(self):
        '''
        Convert the SparseTensor to a scipy.sparse.lil.lil_matrix, which can
        be converted to a numpy.matrix using the command todense().

        Returns
        -------
        y : scipy.sparse.lil.lil_matrix
            The SparseTensor as a scipy.sparse.lil.lil_matrix.

        '''
        assert self.ndim <= 2, \
            'nd sparse arrays are not allowed for d > 2.'

        y = lil_matrix(tuple(self.shape), dtype=float)
        ind = tuple(self.indices.to_list())
        y[ind] = self.data
        return y

    def eval_at_indices(self, ind, dims=None):
        '''
        Evaluate the tensor at indices.

        If dims is None, return
        s(k) = x(indices(k, 1), indices(k, 2), ..., indices(k, d)),
        1 <= k <= self.shape[0].

        If dims is not None, return a partial evaluation: up to a permutation
        (placing the dimensions dims on the left), return
        s(k, i_1, ..., i_d') = x(indices(k, 1), indices(k, 2), ...,
        indices(k, M), i_1, ..., i_d'),
        1 <= k <= self.shape[0], with M = dims.size and d' = self.order - M.

        Parameters
        ----------
        ind : list of numpy.ndarray
            The indices of the tensor.
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices. The default is None,
            indicating that indices refers to all the dimensions.

        Returns
        -------
        evaluations : scipy.sparse.lil.lil_matrix
            The evaluations of the tensor.

        '''
        assert dims is None or np.all(dims == np.arange(self.order)), \
            'Method not implemented.'

        if isinstance(ind, tensap.MultiIndices):
            ind = ind.array

        J = self.indices.array
        loc_J, loc_I = np.nonzero(np.all(ind == J[:, np.newaxis], axis=2))
        evaluations = lil_matrix((ind.shape[0], 1), dtype=float)
        evaluations[loc_I, 0] = self.data[loc_J]
        return evaluations

    def squeeze(self, dims):
        '''
        Remove the singleton dimensions of the tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray, optional
            Dimensions to squeeze. The default is None, indicating all the
            singleton dimensions.

        Returns
        -------
        SparseTensor
            The squeezed tensor.

        '''
        raise NotImplementedError('Method not implemented.')

    def __add__(self, y):
        ind = self.indices.add_indices(y.indices)
        _, rep_x, _ = ind.intersect_indices(self.indices)
        _, rep_y, _ = ind.intersect_indices(y.indices)
        data = np.zeros(ind.cardinal())
        data[rep_x] += self.data
        data[rep_y] += y.data
        return SparseTensor(data, ind, self.shape)

    def __neg__(self):
        raise NotImplementedError('Method not implemented.')

    def __sub__(self, y):
        raise NotImplementedError('Method not implemented.')

    def tensor_vector_product(self, vectors, dims=None):
        '''
        Compute the contraction of the tensor with vectors.

        Compute the contraction of self with each vector contained in the list
        vectors along dimensions specified by dims. The operation is such that
        V[k] is contracted with the dims[k]-th dimension of self.

        Parameters
        ----------
        vectors : numpy.ndarray or list of numpy.ndarray
            The vectors to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        out : SparseTensor
            The tensor after the contractions with the vectors.

        '''
        if dims is None:
            assert isinstance(vectors, list), 'vectors should be a list.'
            assert len(vectors) == self.order, \
                'len(vectors) must be self.order.'
            dims = np.arange(self.order)
        else:
            dims = np.array(dims)
            if not isinstance(vectors, list):
                vectors = [vectors]
            assert len(vectors) == dims.size, \
                'len(vectors) must be equal to dims.size.'

        vectors = [np.ravel(x) for x in vectors]

        out = deepcopy(self)
        not_dims = np.setdiff1d(range(out.order), dims)
        if not_dims.size == 0:
            out.shape = []
        else:
            out.shape = out.shape[not_dims]

        for i in range(dims.size):
            a = out.data * vectors[i][out.indices.array[:, dims[i]]]

            out.indices.array = out.indices.array[:, np.setdiff1d(
                range(out.indices.array.shape[1]), dims[i])]
            out.indices.array = out.indices.array[a != 0, :]

            if out.indices.array.size != 0:
                out.indices.array, ind = np.unique(out.indices.array, axis=0,
                                                   return_inverse=True)

                a = a[a != 0]
                out.data = np.bincount(ind, weights=a)

            dims -= dims > dims[i]

        out.order -= len(vectors)

        if np.size(out.shape) == 0:
            out = out.data

        return out

    def tensor_matrix_product(self, matrices, dims=None):
        '''
        Contract a tensor with matrices.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self, with the indices k given in dims (if provided).

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        out : SparseTensor
            The tensor after the contractions with the matrices.

        '''
        if dims is None:
            assert isinstance(matrices, (list, np.ndarray)), \
                'matrices should be a list or a numpy.ndarray.'
            assert len(matrices) == self.order, \
                'len(matrices) must be self.order.'
            dims = range(self.order)
        else:
            dims = np.atleast_1d(dims)
            if not isinstance(matrices, list):
                matrices = [matrices]
            assert len(matrices) == dims.size, \
                'len(matrices) must be equal to dims.size.'

        k = 0
        out = deepcopy(self)
        for mu in dims:
            perm_dims = np.concatenate(
                ([mu], np.setdiff1d(np.arange(out.order), mu)))
            out = out.transpose(perm_dims)
            if out.order == 1:
                out.shape[1] = 1
            ind = tensap.MultiIndices(
                out.indices.array[out.data != 0, :]).sub2ind(out.shape)
            x1, x2 = np.unravel_index(ind,
                                      [out.shape[0], np.prod(out.shape[1:])],
                                      order='F')
            x2u, x2uind = np.unique(x2, return_inverse=True)
            s = coo_matrix((out.data[out.data != 0],
                            (x1, x2uind)),
                           shape=(out.shape[0], np.max(x2uind)+1))
            a = np.transpose(s.transpose().dot(np.transpose(matrices[k])))
            y1, y2 = np.nonzero(a)
            out.shape[0] = matrices[k].shape[0]
            ind = np.ravel_multi_index((y1, np.reshape(x2u[y2], y1.shape)),
                                       (out.shape[0], np.prod(out.shape[1:])),
                                       order='F')
            out.indices = tensap.MultiIndices.ind2sub(out.shape, ind)
            out.data = np.ravel(a[a != 0])
            out = out.itranspose(perm_dims)
            k += 1
        return out

    def tensor_matrix_product_eval_diag(self, matrices):
        '''
        Evaluate the diagonal of a tensor obtained by contraction with
        matrices.

        Parameters
        ----------
        matrices : list
            The matrices to use in the product.

        Returns
        -------
        SparseTensor
            The diagonal of the contractions of the tensor with the matrices.

        '''
        y = matrices[0][:, self.indices.array[:, 0]]
        for k in np.arange(1, self.order):
            y *= matrices[k][:, self.indices.array[:, k]]
        return np.matmul(y, self.data)

    def transpose(self, dims):
        '''
        Transpose (permute) the dimensions of the tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The new ordering of the dimensions.

        Returns
        -------
        out : SparseTensor
            The transposed (permuted) tensor.

        '''
        out = deepcopy(self)
        out.indices.array = out.indices.array[:, dims]
        out.shape = out.shape[dims]
        return out

    def itranspose(self, dims):
        '''
        Return the inverse transpose (permutation) of the dimensions of the
        tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The original transpose (permutation) indices.

        Returns
        -------
        SparseTensor
            The transposed (permuted) tensor.

        '''
        out = deepcopy(self)
        return out.transpose(np.argsort(dims))

    def reshape(self, shape):
        '''
        Reshape the tensor.

        Parameters
        ----------
        shape : list or numpy.ndarray
            The new shape of the tensor.

        Returns
        -------
        tensor : SparseTensor
            The reshaped tensor.

        '''
        shape = np.array(shape)
        ind = self.indices.sub2ind(self.shape)
        out = deepcopy(self)
        out.indices = tensap.MultiIndices.ind2sub(shape, ind)
        out.shape = shape
        out.order = shape.size
        return out

    def tensor_diagonal_matrix_product(self, matrices, dims=None):
        '''
        Contract a SparseTensor with matrices built from their diagonals.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self, with the indices k given in dims (if provided).

        FIXME: not optimal, does not exploit sparsity.

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The diagonals of the matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        SparseTensor
            The tensor after the contractions with the matrices.

        '''
        if not isinstance(matrices, list):
            matrices = [matrices[:, i] for i in range(np.shape(matrices)[1])]

        if dims is None:
            assert len(matrices) == self.order, \
                'len(matrices) must be self.order.'
            dims = range(self.order)
        else:
            dims = np.array(dims)
            assert len(matrices) == dims.size, \
                'len(matrices) must be equal to dims.size.'

        matrices = [tensap.FullTensor(np.diag(np.reshape(x, [-1])))
                    for x in matrices]
        return self.tensor_matrix_product(matrices, dims)

    def dot(self, y):
        '''
        Return the inner product of two tensors.

        Parameters
        ----------
        y : tensap.Tensor
            The second tensor of the inner products. Must be convertible to
            a SparseTensor.

        Returns
        -------
        numpy.float
            The inner product of the two tensors.

        '''
        if not isinstance(y, SparseTensor):
            try:
                y = y.sparse()
            except Exception:
                raise ValueError('Cannot convert input to SparseTensor.')

        _, ind_x, ind_y = self.indices.intersect_indices(y.indices)
        return np.sum(self.data[ind_x] * y.data[ind_y])

    def __mul__(self, y):
        if not isinstance(y, SparseTensor):
            try:
                y = y.sparse()
            except Exception:
                raise ValueError('Cannot convert input to SparseTensor.')

        _, ind_x, ind_y = self.indices.intersect_indices(y.indices)
        out = deepcopy(self)
        out.data = self.data[ind_x] * y.data[ind_y]
        out.indices.array = self.indices.array[ind_x, :]
        return out

    def norm(self):
        '''
        Compute the canonical norm of the SparseTensor.

        Returns
        -------
        numpy.float
            The norm of the tensor.

        '''
        return np.sqrt(self.dot(self))

    def orth(self, dim):
        '''
        Orthogonalize the tensor.

        Parameters
        ----------
        dim : int
            The dimension of the orthogonal dim-matricization of self.

        Returns
        -------
        SparseTensor
            A tensor whose dim-matricization is an orthogonal matrix
            corresponding to the Q factor of a QR factorization of the
            dim-matricization of self.
        r_matrix : numpy.ndarray
            The R factor.

        '''
        raise NotImplementedError('Method not implemented.')

    def cat(self, y, dims):
        '''
        Concatenate the tensors.

        Concatenates self and y in a tensor z such that:
        z(i_1 ,..., i_d) = x(i_1, ..., i_d) if i_k <= sz[k]-1 for k in dims,
        z(i_1, ..., i_d) = y(i_1-sz[0], ..., i_d-sz[d-1]) if i_k >= sz[k]
        for k in dims,
        z(i_1, ..., i_d) = 0 otherwise, with sz = self.shape and
        dims = range(self.order) if not provided.

        Parameters
        ----------
        y : Tensor
            The second tensor to be concatenaed.
        dims : list or numpy.ndarray, optional
            The dimensions of the concatenation. The default is None,
            indicating all the dimensions.
        Returns
        -------
        SparseTensor
            The concatenated tensors.

        '''
        raise NotImplementedError('Method not implemented.')

    def kron(self, y):
        '''
        Kronecker product of tensors.

        Similar to numpy.kron but for sparse tensors.

        Parameters
        ----------
        y : Tensor
            The second tensor of the Kronecker product.

        Returns
        -------
        SparseTensor
            The tensor resulting from the Kronecker product.

        '''
        raise NotImplementedError('Method not implemented.')

    def dot_with_rank_one_metric(self, y, M):
        '''
        Compute the weighted inner product of two tensors.

        Compute the weighted canonical inner product of self and y,
        where the inner product related to dimension k is weighted by
        M[k]. It is equivalent to
        self.dot(y.tensor_matrix_product(M)),
        but can be much faster.

        Parameters
        ----------
        y : Tensor
            The second tensor of the inner product.
        M : list or numpy.ndarray or FullTensor
            The weight matrix.

        Returns
        -------
        numpy.float
            The weighted inner product.

        '''
        s = y.tensor_matrix_product(M)
        return self.dot(s)

    def tensordot_matrix_product_except_dim(self, y, M, dim):
        '''
        Particular type of contraction.

        Compute a special contraction of two tensors self, y, a list of
        matrices M and a particular dimension dim. Note that dim must
        be a scalar, while M must be a list array with x.self.order
        elements.

        Parameters
        ----------
        y : Tensor
            The second tensor of the contraction.
        M : list
            The list of matrices of the contraction.
        dim : int
            The excluded dimension.

        Returns
        -------
        numpy.ndarray
            The result of the contraction.

        '''
        # dims = np.setdiff1d(np.arange(self.order), dim)
        # s = y.tensor_matrix_product(M[dims], dims)
        # return self.tensordot(s, dims, dims)
        raise NotImplementedError('Method not implemented.')

    def eval_diag(self, dims=None):
        '''
        Extract the diagonal of the tensor.

        The tensor must be such that self.shape[mu] = n for all mu (in dims if
        provided).

        Parameters
        ----------
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices of the diagonal. The
            default is None,indicating that indices refers to all the
            dimensions.

        Returns
        -------
        data : numpy.ndarray
            The evaluations of the diagonal of the tensor.

        '''
        if dims is None:
            dims = np.arange(self.order)
        else:
            dims = np.atleast_1d(dims)

        if dims.size == 1:
            data = self
        else:
            assert np.all([self.shape[x] == self.shape[dims[0]] for
                           x in dims]),\
             'The shapes of the tensor in dimensions dims should be equal.'
            ind = np.repeat(np.reshape(np.arange(self.shape[0]), [-1, 1]),
                            dims.size, 1)
            data = self.eval_at_indices(ind, dims)
        return data

    def sub_tensor(self, *indices):
        '''
        Extract a subtensor of the tensor.

        The result is a tensor s of shape
        len(indices[0]), ..., len(indices[self.order-1]),
        such that
        s(k1,...,kd) = x(indices[0][k1], ..., indices[self.order-1][kd]).

        Example: x.subTensor([1, 2], ':', [2, 5, 6]) returns a tensor with
        shape [2, self.shape[1], 3].

        Parameters
        ----------
        *indices : list
            The indices to extract in each dimension. ':' indicates all the
            indices.

        Returns
        -------
        SparseTensor
            The subtensor.

        '''
        raise NotImplementedError('Method not implemented.')



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module sub_functional_basis.

'''

from copy import deepcopy
import numpy as np
import tensap


class SubFunctionalBasis(tensap.FunctionalBasis):
    '''
    Class SubFunctionalBasis.

    Attributes
    ----------
    underlying_basis : tensap.FunctionalBasis
        The underlying basis.

    basis : numpy.ndarray
        Array of shape (n, m), where n is the number of elements in
        underlying_basis, which defines a set of m basis functions in the space
        generated by underlying_basis.

    '''

    def __init__(self, underlying_basis=None, basis=None):
        '''
        Constructor for the class SubFunctionalBasis.

        Parameters
        ----------
        underlying_basis : tensap.FunctionalBasis
            The underlying basis.
        basis : numpy.ndarray
            Array of shape (n, m), where n is the number of elements in
            underlying_basis, which defines a set of m basis functions in the
            space generated by underlying_basis.

        Returns
        -------
        None.

        '''
        tensap.FunctionalBasis.__init__(self)
        self.underlying_basis = deepcopy(underlying_basis)
        self.basis = None
        if underlying_basis is not None:
            if basis is None:
                self.basis = np.zeros((self.underlying_basis.cardinal(), 0))
            else:
                self.basis = basis

            self.measure = underlying_basis.measure

            if underlying_basis.is_orthonormal and \
                np.linalg.norm(np.matmul(np.transpose(basis), basis) -
                               np.eye(basis.shape[1])) / \
                    np.sqrt(basis.shape[1]) < 1e-15:
                self.is_orthonormal = True

    def storage(self):
        return self.basis.size

    def cardinal(self):
        return self.basis.shape[1]

    def ndim(self):
        return self.underlying_basis.ndim()

    def domain(self):
        return self.underlying_basis.domain()

    def orthonormalize(self):
        '''
        Orthonormalize the SubFunctionalBasis.

        Returns
        -------
        tensap.SubFunctionalBasis
            The orthonormalized SubFunctionalBasis.

        '''
        assert self.underlying_basis.is_orthonormal, \
            'Method not implemented.'

        new_basis = np.array(self.basis)
        new_basis[:, 0] /= np.linalg.norm(new_basis[:, 0])
        for i in np.arange(1, new_basis.shape[1]):
            new_basis[:, i] -= np.matmul(new_basis[:, :i], np.matmul(
                np.transpose(new_basis[:, :i]), new_basis[:, i]))
            new_basis[:, i] /= np.linalg.norm(new_basis[:, i])

        return SubFunctionalBasis(self.underlying_basis, new_basis)

    def eval(self, x, indices=None):
        if indices is None:
            indices = np.arange(self.cardinal())

        return np.matmul(self.underlying_basis.eval(x), self.basis[:, indices])

    def eval_derivative(self, n, x):
        '''
        Evaluate the k-th order derivative of the functions of the basis at the
        points x.

        Parameters
        ----------
        k : int
            The order of the derivative.
        x : list or numpy.ndarray
            The points at which the k-th derivative of the basis functions are
            to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the k-th derivative of the functions of
            self.basis of degrees in self.indices at points x.

        '''
        return np.matmul(self.underlying_basis.eval_derivative(x, n),
                         self.basis)

    def derivative(self, n):
        '''
        Compute the k-th order derivative of the functions of the basis
        projected on itself.

        Parameters
        ----------
        k : int
            The order of the derivative.

        Returns
        -------
        tensap.SubFunctionalBasis
            The k-th order derivative of the functions of the basis projected
            on itself.

        '''
        d_f = deepcopy(self)
        d_f.underlying_basis = self.underlying_basis.derivative(n)
        return d_f

    def mean(self):
        return np.matmul(self.underlying_basis.mean(), self.basis)
